from nose.tools import assert_equal, assert_not_equal

from whoosh import spelling
from whoosh.filedb.filestore import RamStorage


def test_spelling():
    st = RamStorage()
    
    sp = spelling.SpellChecker(st, mingram=2)
    
    wordlist = ["render", "animation", "animate", "shader",
                "shading", "zebra", "koala", "lamppost",
                "ready", "kismet", "reaction", "page",
                "delete", "quick", "brown", "fox", "jumped",
                "over", "lazy", "dog", "wicked", "erase",
                "red", "team", "yellow", "under", "interest",
                "open", "print", "acrid", "sear", "deaf",
                "feed", "grow", "heal", "jolly", "kilt",
                "low", "zone", "xylophone", "crown",
                "vale", "brown", "neat", "meat", "reduction",
                "blunder", "preaction"]
    
    sp.add_words([unicode(w) for w in wordlist])
    
    sugs = sp.suggest(u"reoction")
    assert_not_equal(len(sugs), 0)
    assert_equal(sugs, [u"reaction", u"reduction", u"preaction"])
    
def test_suggestionsandscores():
    st = RamStorage()
    sp = spelling.SpellChecker(st, mingram=2)
    
    words = [("alfa", 10), ("bravo", 9), ("charlie", 8), ("delta", 7),
             ("echo", 6), ("foxtrot", 5), ("golf", 4), ("hotel", 3),
             ("india", 2), ("juliet", 1)]
    sp.add_scored_words((unicode(w), s) for w, s in words)
    
    from whoosh.scoring import Frequency
    sugs = sp.suggestions_and_scores(u"alpha", weighting=Frequency())
    assert_equal(sugs, [(u"alfa", 10, 3.0), (u"charlie", 8, 1.0)])

def test_minscore():
    st = RamStorage()
    sp = spelling.SpellChecker(st, mingram=2, minscore=2.0)
    
    sp.add_words([u'charm', u'amour'])
    
    sugs = sp.suggest(u"armor")
    assert_equal(sugs, [u'charm'])







from __future__ import with_statement

from nose.tools import assert_equal, assert_not_equal, assert_raises

from whoosh import analysis, fields, formats, qparser, query, searching
from whoosh.filedb.filestore import RamStorage
from whoosh.util import permutations


def test_score_retrieval():
    schema = fields.Schema(title=fields.TEXT(stored=True),
                           content=fields.TEXT(stored=True))
    storage = RamStorage()
    ix = storage.create_index(schema)
    writer = ix.writer()
    writer.add_document(title=u"Miss Mary",
                        content=u"Mary had a little white lamb its fleece was white as snow")
    writer.add_document(title=u"Snow White",
                        content=u"Snow white lived in the forest with seven dwarfs")
    writer.commit()
    
    with ix.searcher() as s:
        results = s.search(query.Term("content", "white"))
        assert_equal(len(results), 2)
        assert_equal(results[0]['title'], u"Miss Mary")
        assert_equal(results[1]['title'], u"Snow White")
        assert_not_equal(results.score(0), None)
        assert_not_equal(results.score(0), 0)
        assert_not_equal(results.score(0), 1)

def test_resultcopy():
    schema = fields.Schema(a=fields.TEXT(stored=True))
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(a=u"alfa bravo charlie")
    w.add_document(a=u"bravo charlie delta")
    w.add_document(a=u"charlie delta echo")
    w.add_document(a=u"delta echo foxtrot")
    w.commit()
    
    with ix.searcher() as s:
        r = s.search(qparser.QueryParser("a", None).parse(u"charlie"))
        assert_equal(len(r), 3)
        rcopy = r.copy()
        assert_equal(r.top_n, rcopy.top_n)
    
def test_resultslength():
    schema = fields.Schema(id=fields.ID(stored=True),
                           value=fields.TEXT)
    ix = RamStorage().create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1", value=u"alfa alfa alfa alfa alfa")
    w.add_document(id=u"2", value=u"alfa alfa alfa alfa")
    w.add_document(id=u"3", value=u"alfa alfa alfa")
    w.add_document(id=u"4", value=u"alfa alfa")
    w.add_document(id=u"5", value=u"alfa")
    w.add_document(id=u"6", value=u"bravo")
    w.commit()
    
    with ix.searcher() as s:
        r = s.search(query.Term("value", u"alfa"), limit=3)
        assert_equal(len(r), 5)
        assert_equal(r.scored_length(), 3)
        assert_equal(r[10:], [])

def test_combine():
    schema = fields.Schema(id=fields.ID(stored=True),
                           value=fields.TEXT)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=u"1", value=u"alfa bravo charlie all")
    w.add_document(id=u"2", value=u"bravo charlie echo all")
    w.add_document(id=u"3", value=u"charlie echo foxtrot all")
    w.add_document(id=u"4", value=u"echo foxtrot india all")
    w.add_document(id=u"5", value=u"foxtrot india juliet all")
    w.add_document(id=u"6", value=u"india juliet alfa all")
    w.add_document(id=u"7", value=u"juliet alfa bravo all")
    w.add_document(id=u"8", value=u"charlie charlie charlie all")
    w.commit()
    
    with ix.searcher() as s:
        def idsof(r):
            return "".join(hit["id"] for hit in r)
        
        def check(r1, methodname, r2, ids):
            getattr(r1, methodname)(r2)
            assert_equal(idsof(r1), ids)
        
        def rfor(t):
            return s.search(query.Term("value", t))
        
        assert_equal(idsof(rfor(u"foxtrot")), "345")
        check(rfor(u"foxtrot"), "extend", rfor("charlie"), "345812")
        check(rfor(u"foxtrot"), "filter", rfor("juliet"), "5")
        check(rfor(u"charlie"), "filter", rfor("foxtrot"), "3")
        check(rfor(u"all"), "filter", rfor("foxtrot"), "345")
        check(rfor(u"all"), "upgrade", rfor("india"), "45612378")
        check(rfor(u"charlie"), "upgrade_and_extend", rfor("echo"), "23814")

def test_results_filter():
    schema = fields.Schema(id=fields.STORED, words=fields.KEYWORD(stored=True))
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id="1", words=u"bravo top")
    w.add_document(id="2", words=u"alfa top")
    w.add_document(id="3", words=u"alfa top")
    w.add_document(id="4", words=u"alfa bottom")
    w.add_document(id="5", words=u"bravo bottom")
    w.add_document(id="6", words=u"charlie bottom")
    w.add_document(id="7", words=u"charlie bottom")
    w.commit()
    
    with ix.searcher() as s:
        def check(r, target):
            result = "".join(s.stored_fields(d)["id"] for d in r.docs())
            assert_equal(result, target)
        
        r = s.search(query.Term("words", u"alfa"))
        r.filter(s.search(query.Term("words", u"bottom")))
        check(r, "4")
    
def test_pages():
    from whoosh.scoring import Frequency
    
    schema = fields.Schema(id=fields.ID(stored=True), c=fields.TEXT)
    ix = RamStorage().create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1", c=u"alfa alfa alfa alfa alfa alfa")
    w.add_document(id=u"2", c=u"alfa alfa alfa alfa alfa")
    w.add_document(id=u"3", c=u"alfa alfa alfa alfa")
    w.add_document(id=u"4", c=u"alfa alfa alfa")
    w.add_document(id=u"5", c=u"alfa alfa")
    w.add_document(id=u"6", c=u"alfa")
    w.commit()
    
    with ix.searcher(weighting=Frequency) as s:
        q = query.Term("c", u"alfa")
        r = s.search(q)
        assert_equal([d["id"] for d in r], ["1", "2", "3", "4", "5", "6"])
        r = s.search_page(q, 2, pagelen=2)
        assert_equal([d["id"] for d in r], ["3", "4"])
        
        r = s.search_page(q, 2, pagelen=4)
        assert_equal(r.total, 6)
        assert_equal(r.pagenum, 2)
        assert_equal(r.pagelen, 2)

def test_extra_slice():
    schema = fields.Schema(key=fields.ID(stored=True))
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    for char in u"abcdefghijklmnopqrstuvwxyz":
        w.add_document(key=char)
    w.commit()
    
    with ix.searcher() as s:
        r = s.search(query.Every(), limit=5)
        assert_equal(r[6:7], [])

def test_page_counts():
    from whoosh.scoring import Frequency
    
    schema = fields.Schema(id=fields.ID(stored=True))
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    for i in xrange(10):
        w.add_document(id=unicode(i))
    w.commit()
    
    with ix.searcher(weighting=Frequency) as s:
        q = query.Every("id")
        
        r = s.search(q)
        assert_equal(len(r), 10)
        
        assert_raises(ValueError, s.search_page, q, 0)
        
        r = s.search_page(q, 1, 5)
        assert_equal(len(r), 10)
        assert_equal(r.pagecount, 2)
        
        r = s.search_page(q, 1, 5)
        assert_equal(len(r), 10)
        assert_equal(r.pagecount, 2)
        
        r = s.search_page(q, 2, 5)
        assert_equal(len(r), 10)
        assert_equal(r.pagecount, 2)
        assert_equal(r.pagenum, 2)
        
        r = s.search_page(q, 1, 10)
        assert_equal(len(r), 10)
        assert_equal(r.pagecount, 1)
        assert_equal(r.pagenum, 1)

def test_resultspage():
    schema = fields.Schema(id=fields.STORED, content=fields.TEXT)
    ix = RamStorage().create_index(schema)
    
    domain = ("alfa", "bravo", "bravo", "charlie", "delta")
    w = ix.writer()
    for i, lst in enumerate(permutations(domain, 3)):
        w.add_document(id=unicode(i), content=u" ".join(lst))
    w.commit()
    
    with ix.searcher() as s:
        q = query.Term("content", u"bravo")
        r = s.search(q, limit=10)
        tops = list(r)
        
        rp = s.search_page(q, 1, pagelen=5)
        assert_equal(list(rp), tops[0:5])
        assert_equal(rp[10:], [])
        
        rp = s.search_page(q, 2, pagelen=5)
        assert_equal(list(rp), tops[5:10])
        
        rp = s.search_page(q, 1, pagelen=10)
        assert_equal(len(rp), 54)
        assert_equal(rp.pagecount, 6)
        rp = s.search_page(q, 6, pagelen=10)
        assert_equal(len(list(rp)), 4)
        assert rp.is_last_page()
        
        assert_raises(ValueError, s.search_page, q, 0)
        assert_raises(ValueError, s.search_page, q, 7)
        
        rp = s.search_page(query.Term("content", "glonk"), 1)
        assert_equal(len(rp), 0)
        assert rp.is_last_page()

def test_snippets():
    from whoosh import highlight
    
    ana = analysis.StemmingAnalyzer()
    schema = fields.Schema(text=fields.TEXT(stored=True, analyzer=ana))
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(text=u"Lay out the rough animation by creating the important poses where they occur on the timeline.")
    w.add_document(text=u"Set key frames on everything that's key-able. This is for control and predictability: you don't want to accidentally leave something un-keyed. This is also much faster than selecting the parameters to key.")
    w.add_document(text=u"Use constant (straight) or sometimes linear transitions between keyframes in the channel editor. This makes the character jump between poses.")
    w.add_document(text=u"Keying everything gives quick, immediate results, but it can become difficult to tweak the animation later, especially for complex characters.")
    w.add_document(text=u"Copy the current pose to create the next one: pose the character, key everything, then copy the keyframe in the playbar to another frame, and key everything at that frame.")
    w.commit()
    
    target = ["Set KEY frames on everything that's KEY-able. This is for control and predictability...leave something un-KEYED. This is also much faster than selecting...parameters to KEY",
              "next one: pose the character, KEY everything, then copy...playbar to another frame, and KEY everything at that frame",
              "KEYING everything gives quick, immediate results"]
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("text", ix.schema)
        q = qp.parse(u"key")
        r = s.search(q)
        r.formatter = highlight.UppercaseFormatter()
        
        assert_equal([hit.highlights("text") for hit in r], target)
        
def test_keyterms():
    ana = analysis.StandardAnalyzer()
    vectorformat = formats.Frequency(ana)
    schema = fields.Schema(path=fields.ID,
                           content=fields.TEXT(analyzer=ana,
                                               vector=vectorformat))
    st = RamStorage()
    ix = st.create_index(schema)
    w = ix.writer()
    w.add_document(path=u"a",content=u"This is some generic content")
    w.add_document(path=u"b",content=u"This is some distinctive content")
    w.commit()
    
    with ix.searcher() as s:
        docnum = s.document_number(path=u"b")
        keyterms = list(s.key_terms([docnum], "content"))
        assert len(keyterms) > 0
        assert_equal(keyterms[0][0], "distinctive")
        
        r = s.search(query.Term("path", u"b"))
        keyterms2 = list(r.key_terms("content"))
        assert len(keyterms2) > 0
        assert_equal(keyterms2[0][0], "distinctive")
    
def test_lengths():
    schema = fields.Schema(id=fields.STORED, text=fields.TEXT)
    ix = RamStorage().create_index(schema)
    
    w = ix.writer()
    w.add_document(id=1, text=u"alfa bravo charlie delta echo")
    w.add_document(id=2, text=u"bravo charlie delta echo foxtrot")
    w.add_document(id=3, text=u"charlie needle echo foxtrot golf")
    w.add_document(id=4, text=u"delta echo foxtrot golf hotel")
    w.add_document(id=5, text=u"echo needle needle hotel india")
    w.add_document(id=6, text=u"foxtrot golf hotel india juliet")
    w.add_document(id=7, text=u"golf needle india juliet kilo")
    w.add_document(id=8, text=u"hotel india juliet needle lima")
    w.commit()
    
    with ix.searcher() as s:
        q = query.Or([query.Term("text", u"needle"), query.Term("text", u"charlie")])
        r = s.search(q, limit=2)
        assert_equal(r.has_exact_length(), False)
        assert_equal(r.estimated_length(), 7)
        assert_equal(r.estimated_min_length(), 3)
        assert_equal(r.scored_length(), 2)
        assert_equal(len(r), 6)

def test_lengths2():
    schema = fields.Schema(text=fields.TEXT(stored=True))
    ix = RamStorage().create_index(schema)
    count = 0
    for _ in xrange(3):
        w = ix.writer()
        for ls in permutations(u"alfa bravo charlie".split()):
            if "bravo" in ls and "charlie" in ls:
                count += 1
            w.add_document(text=u" ".join(ls))
        w.commit(merge=False)
    
    with ix.searcher() as s:
        q = query.Or([query.Term("text", u"bravo"), query.Term("text", u"charlie")])
        r = s.search(q, limit=None)
        assert_equal(len(r), count)
        
        r = s.search(q, limit=3)
        assert_equal(len(r), count)

def test_stability():
    schema = fields.Schema(text=fields.TEXT)
    ix = RamStorage().create_index(schema)
    domain = u"alfa bravo charlie delta".split()
    w = ix.writer()
    for ls in permutations(domain, 3):
        w.add_document(text=u" ".join(ls))
    w.commit()
    
    with ix.searcher() as s:
        q = query.Term("text", u"bravo")
        last = []
        for i in xrange(s.doc_frequency("text", u"bravo")):
            # Only un-optimized results are stable
            r = s.search(q, limit=i + 1, optimize=False)
            docnums = [hit.docnum for hit in r]
            assert_equal(docnums[:-1], last)
            last = docnums



















        
    


from nose.tools import assert_equal, assert_not_equal

from whoosh import fields, scoring
from whoosh.filedb.filestore import RamStorage
from whoosh.qparser import QueryParser
from whoosh.query import *
from whoosh.spans import *


def test_all_terms():
    q = QueryParser("a", None).parse(u'hello b:there c:"my friend"')
    ts = set()
    q.all_terms(ts, phrases=False)
    assert_equal(sorted(ts), [("a", "hello"), ("b", "there")])
    ts = set()
    q.all_terms(ts, phrases=True)
    assert_equal(sorted(ts), [("a", "hello"), ("b", "there"), ("c", "friend"), ("c", "my")])

def test_existing_terms():
    s = fields.Schema(key=fields.ID, value=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(s)
    
    w = ix.writer()
    w.add_document(key=u"a", value=u"alfa bravo charlie delta echo")
    w.add_document(key=u"b", value=u"foxtrot golf hotel india juliet")
    w.commit()
    
    r = ix.reader()
    q = QueryParser("value", None).parse(u'alfa hotel tango "sierra bravo"')
    
    ts = q.existing_terms(r, phrases=False)
    assert_equal(sorted(ts), [("value", "alfa"), ("value", "hotel")])
    
    ts = q.existing_terms(r)
    assert_equal(sorted(ts), [("value", "alfa"), ("value", "bravo"), ("value", "hotel")])
    
    ts = set()
    q.existing_terms(r, ts, reverse=True)
    assert_equal(sorted(ts), [("value", "sierra"), ("value", "tango")])

def test_replace():
    q = And([Or([Term("a", "b"), Term("b", "c")], boost=1.2), Variations("a", "b", boost=2.0)])
    q = q.replace("b", "BB")
    assert_equal(q, And([Or([Term("a", "BB"), Term("b", "c")], boost=1.2),
                         Variations("a", "BB", boost=2.0)]))

def test_apply():
    def visit(q):
        if isinstance(q, (Term, Variations, FuzzyTerm)):
            q.text = q.text.upper()
            return q
        return q.apply(visit)
    
    before = And([Not(Term("a", u"b")), Variations("a", u"c"), Not(FuzzyTerm("a", u"d"))])
    after = visit(before)
    assert_equal(after, And([Not(Term("a", u"B")), Variations("a", u"C"),
                             Not(FuzzyTerm("a", u"D"))]))
    
    def term2var(q):
        if isinstance(q, Term):
            return Variations(q.fieldname, q.text)
        else:
            return q.apply(term2var)

    q = And([Term("f", "alfa"), Or([Term("f", "bravo"), Not(Term("f", "charlie"))])])
    q = term2var(q)
    assert_equal(q, And([Variations('f', 'alfa'),
                         Or([Variations('f', 'bravo'), Not(Variations('f', 'charlie'))])]))

def test_accept():
    def boost_phrases(q):
        if isinstance(q, Phrase):
            q.boost *= 2.0
        return q
    
    before = And([Term("a", u"b"), Or([Term("c", u"d"), Phrase("a", [u"e", u"f"])]),
                  Phrase("a", [u"g", u"h"], boost=0.25)])
    after = before.accept(boost_phrases)
    assert_equal(after, And([Term("a", u"b"),
                             Or([Term("c", u"d"), Phrase("a", [u"e", u"f"], boost=2.0)]),
                             Phrase("a", [u"g", u"h"], boost=0.5)]))

    before = Phrase("a", [u"b", u"c"], boost=2.5)
    after = before.accept(boost_phrases)
    assert_equal(after, Phrase("a", [u"b", u"c"], boost=5.0))

def test_simplify():
    s = fields.Schema(k=fields.ID, v=fields.TEXT)
    ix = RamStorage().create_index(s)
    
    w = ix.writer()
    w.add_document(k=u"1", v=u"aardvark apple allan alfa bear bee")
    w.add_document(k=u"2", v=u"brie glue geewhiz goop julia")
    w.commit()
    
    r = ix.reader()
    q1 = And([Prefix("v", "b", boost=2.0), Term("v", "juliet")])
    q2 = And([Or([Term('v', u'bear', boost=2.0), Term('v', u'bee', boost=2.0),
                  Term('v', u'brie', boost=2.0)]), Term('v', 'juliet')])
    assert_equal(q1.simplify(r), q2)
    
def test_merge_ranges():
    q = And([TermRange("f1", u"a", None), TermRange("f1", None, u"z")])
    assert_equal(q.normalize(), TermRange("f1", u"a", u"z"))
    
    q = And([NumericRange("f1", None, u"aaaaa"), NumericRange("f1", u"zzzzz", None)])
    assert_equal(q.normalize(), q)
    
    q = And([TermRange("f1", u"a", u"z"), TermRange("f1", "b", "x")])
    assert_equal(q.normalize(), TermRange("f1", u"a", u"z"))
    
    q = And([TermRange("f1", u"a", u"m"), TermRange("f1", u"f", u"q")])
    assert_equal(q.normalize(), TermRange("f1", u"f", u"m"))
    
    q = Or([TermRange("f1", u"a", u"m"), TermRange("f1", u"f", u"q")])
    assert_equal(q.normalize(), TermRange("f1", u"a", u"q"))
    
    q = Or([TermRange("f1", u"m", None), TermRange("f1", None, u"n")])
    assert_equal(q.normalize(), Every("f1"))
    
    q = And([Every("f1"), Term("f1", "a"), Variations("f1", "b")])
    assert_equal(q.normalize(), Every("f1"))
    
    q = Or([Term("f1", u"q"), TermRange("f1", u"m", None), TermRange("f1", None, u"n")])
    assert_equal(q.normalize(), Every("f1"))
    
    q = And([Or([Term("f1", u"a"), Term("f1", u"b")]), Every("f1")])
    assert_equal(q.normalize(), Every("f1"))
    
    q = And([Term("f1", u"a"), And([Or([Every("f1")])])])
    assert_equal(q.normalize(), Every("f1"))
    
def test_normalize_compound():
    def oq():
        return Or([Term("a", u"a"), Term("a", u"b")])
    def nq(level):
        if level == 0:
            return oq()
        else:
            return Or([nq(level-1), nq(level-1), nq(level-1)])
    
    q = nq(7)
    q = q.normalize()
    assert_equal(q, Or([Term("a", u"a"), Term("a", u"b")]))

def test_duplicates():
    q = And([Term("a", u"b"), Term("a", u"b")])
    assert_equal(q.normalize(), Term("a", u"b"))
    
    q = And([Prefix("a", u"b"), Prefix("a", u"b")])
    assert_equal(q.normalize(), Prefix("a", u"b"))
    
    q = And([Variations("a", u"b"), And([Variations("a", u"b"), Term("a", u"b")])])
    assert_equal(q.normalize(), And([Variations("a", u"b"), Term("a", u"b")]))
    
    q = And([Term("a", u"b"), Prefix("a", u"b"), Term("a", u"b", boost=1.1)])
    assert_equal(q.normalize(), q)
    
    # Wildcard without * or ? normalizes to Term
    q = And([Wildcard("a", u"b"), And([Wildcard("a", u"b"), Term("a", u"b")])])
    assert_equal(q.normalize(), Term("a", u"b"))

def test_query_copy_hash():
    def do(q1, q2):
        q1a = q1.copy()
        assert_equal(q1, q1a)
        assert_equal(hash(q1), hash(q1a))
        assert_not_equal(q1, q2)
        
    do(Term("a", u"b", boost=1.1), Term("a", u"b", boost=1.5))
    do(And([Term("a", u"b"), Term("c", u"d")], boost=1.1),
       And([Term("a", u"b"), Term("c", u"d")], boost=1.5))
    do(Or([Term("a", u"b", boost=1.1), Term("c", u"d")]),
       Or([Term("a", u"b", boost=1.8), Term("c", u"d")], boost=1.5))
    do(DisjunctionMax([Term("a", u"b", boost=1.8), Term("c", u"d")]),
       DisjunctionMax([Term("a", u"b", boost=1.1), Term("c", u"d")], boost=1.5))
    do(Not(Term("a", u"b", boost=1.1)), Not(Term("a", u"b", boost=1.5)))
    do(Prefix("a", u"b", boost=1.1), Prefix("a", u"b", boost=1.5))
    do(Wildcard("a", u"b*x?", boost=1.1), Wildcard("a", u"b*x?", boost=1.5))
    do(FuzzyTerm("a", u"b", constantscore=True),
       FuzzyTerm("a", u"b", constantscore=False))
    do(FuzzyTerm("a", u"b", boost=1.1), FuzzyTerm("a", u"b", boost=1.5))
    do(TermRange("a", u"b", u"c"), TermRange("a", u"b", u"d"))
    do(TermRange("a", None, u"c"), TermRange("a", None, None))
    do(TermRange("a", u"b", u"c", boost=1.1),
       TermRange("a", u"b", u"c", boost=1.5))
    do(TermRange("a", u"b", u"c", constantscore=True),
       TermRange("a", u"b", u"c", constantscore=False))
    do(NumericRange("a", 1, 5), NumericRange("a", 1, 6))
    do(NumericRange("a", None, 5), NumericRange("a", None, None))
    do(NumericRange("a", 3, 6, boost=1.1), NumericRange("a", 3, 6, boost=1.5))
    do(NumericRange("a", 3, 6, constantscore=True),
       NumericRange("a", 3, 6, constantscore=False))
    # do(DateRange)
    do(Variations("a", u"render"), Variations("a", u"renders"))
    do(Variations("a", u"render", boost=1.1),
       Variations("a", u"renders", boost=1.5))
    do(Phrase("a", [u"b", u"c", u"d"]), Phrase("a", [u"b", u"c", u"e"]))
    do(Phrase("a", [u"b", u"c", u"d"], boost=1.1),
       Phrase("a", [u"b", u"c", u"d"], boost=1.5))
    do(Phrase("a", [u"b", u"c", u"d"], slop=1),
       Phrase("a", [u"b", u"c", u"d"], slop=2))
    # do(Ordered)
    do(Every(), Every("a"))
    do(Every("a"), Every("b"))
    do(Every("a", boost=1.1), Every("a", boost=1.5))
    do(NullQuery, Term("a", u"b"))
    do(ConstantScoreQuery(Term("a", u"b")), ConstantScoreQuery(Term("a", u"c")))
    do(ConstantScoreQuery(Term("a", u"b"), score=2.0),
       ConstantScoreQuery(Term("a", u"c"), score=2.1))
    do(WeightingQuery(Term("a", u"b"), scoring.Frequency()),
       WeightingQuery(Term("a", u"c"), scoring.Frequency()))
    do(Require(Term("a", u"b"), Term("c", u"d")),
       Require(Term("a", u"b", boost=1.1), Term("c", u"d")))
    # do(Require)
    # do(AndMaybe)
    # do(AndNot)
    # do(Otherwise)
    
    do(SpanFirst(Term("a", u"b"), limit=1), SpanFirst(Term("a", u"b"), limit=2))
    do(SpanNear(Term("a", u"b"), Term("c", u"d")),
       SpanNear(Term("a", u"b"), Term("c", u"e")))
    do(SpanNear(Term("a", u"b"), Term("c", u"d"), slop=1),
       SpanNear(Term("a", u"b"), Term("c", u"d"), slop=2))
    do(SpanNear(Term("a", u"b"), Term("c", u"d"), mindist=1),
       SpanNear(Term("a", u"b"), Term("c", u"d"), mindist=2))
    do(SpanNear(Term("a", u"b"), Term("c", u"d"), ordered=True),
       SpanNear(Term("a", u"b"), Term("c", u"d"), ordered=False))
    do(SpanNot(Term("a", u"b"), Term("a", u"c")),
       SpanNot(Term("a", u"b"), Term("a", u"d")))
    do(SpanOr([Term("a", u"b"), Term("a", u"c"), Term("a", u"d")]),
       SpanOr([Term("a", u"b"), Term("a", u"c"), Term("a", u"e")]))
    do(SpanContains(Term("a", u"b"), Term("a", u"c")),
       SpanContains(Term("a", u"b"), Term("a", u"d")))
    # do(SpanBefore)
    # do(SpanCondition)

def test_requires():
    a = Term("f", u"a")
    b = Term("f", u"b")
    assert_equal(And([a, b]).requires(), set([a, b]))
    assert_equal(Or([a, b]).requires(), set())
    assert_equal(AndMaybe(a, b).requires(), set([a]))
    assert_equal(a.requires(), set([a]))









import shutil
import tempfile
from functools import wraps

from whoosh.filedb.filestore import FileStorage


class TempStorage(object):
    def __init__(self, basename='', parentdir=None, suppress=frozenset(),
                 keepdir=False):
        self.basename = basename
        self.parentdir = parentdir
        self.suppress = suppress
        self.keepdir = keepdir
        self.dir = None
    
    def __enter__(self):
        self.dir = tempfile.mkdtemp(prefix=self.basename, suffix=".tmpix",
                                    dir=self.parentdir)
        return FileStorage(self.dir)
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.keepdir:
            try:
                shutil.rmtree(self.dir)
            except OSError, e:
                print "Can't remove temp dir: " + str(e)
        
        if exc_type is not None:
            if self.keepdir:
                print "Temp dir=", self.dir
            if exc_type not in self.suppress:
                return False


class TempIndex(TempStorage):
    def __init__(self, schema, ixname='', **kwargs):
        super(TempIndex, self).__init__(basename=ixname, **kwargs)
        self.schema = schema

    def __enter__(self):
        fstore = super(TempIndex, self).__enter__()
        return fstore.create_index(self.schema, indexname=self.basename)


def skip_if(cond):
    """A Nose test decorator that skips the decorated test if the given
    function returns True at runtime.
    """
    
    def decorating_function(testfn):
        @wraps(testfn)
        def wrapper(*args, **kwargs):
            if cond():
                from nose.plugins.skip import SkipTest
                raise SkipTest
            else:
                return testfn(*args, **kwargs)
        
        return wrapper
    return decorating_function
    

def skip_if_unavailable(modulename):
    """A Nose test decorator that only runs the decorated test if a module
    can be imported::
    
        @skip_if_unavailable("multiprocessing")
        def test_mp():
    
    Raises ``SkipTest`` if the module cannot be imported.
    """
    
    def cantimport():
        try:
            __import__(modulename)
        except ImportError:
            return True
        else:
            return False
        
    return skip_if(cantimport)






from __future__ import with_statement

from nose.tools import assert_equal

from whoosh import analysis, fields, reading
from whoosh.filedb.filestore import RamStorage

def _create_index():
    s = fields.Schema(f1 = fields.KEYWORD(stored = True),
                      f2 = fields.KEYWORD,
                      f3 = fields.KEYWORD)
    st = RamStorage()
    ix = st.create_index(s)
    return ix

def _one_segment_index():
    ix = _create_index()
    w = ix.writer()
    w.add_document(f1 = u"A B C", f2 = u"1 2 3", f3 = u"X Y Z")
    w.add_document(f1 = u"D E F", f2 = u"4 5 6", f3 = u"Q R S")
    w.add_document(f1 = u"A E C", f2 = u"1 4 6", f3 = u"X Q S")
    w.add_document(f1 = u"A A A", f2 = u"2 3 5", f3 = u"Y R Z")
    w.add_document(f1 = u"A B", f2 = u"1 2", f3 = u"X Y")
    w.commit()
    
    return ix

def _multi_segment_index():
    ix = _create_index()
    w = ix.writer()
    w.add_document(f1 = u"A B C", f2 = u"1 2 3", f3 = u"X Y Z")
    w.add_document(f1 = u"D E F", f2 = u"4 5 6", f3 = u"Q R S")
    w.commit()
    
    w = ix.writer()
    w.add_document(f1 = u"A E C", f2 = u"1 4 6", f3 = u"X Q S")
    w.add_document(f1 = u"A A A", f2 = u"2 3 5", f3 = u"Y R Z")
    w.commit(merge=False)
    
    w = ix.writer()
    w.add_document(f1 = u"A B", f2 = u"1 2", f3 = u"X Y")
    w.commit(merge=False)
    
    return ix

def test_readers():
    target = [("f1", u'A', 4, 6), ("f1", u'B', 2, 2), ("f1", u'C', 2, 2),
              ("f1", u'D', 1, 1), ("f1", u'E', 2, 2), ("f1", u'F', 1, 1),
              ("f2", u'1', 3, 3), ("f2", u'2', 3, 3), ("f2", u'3', 2, 2),
              ("f2", u'4', 2, 2), ("f2", u'5', 2, 2), ("f2", u'6', 2, 2),
              ("f3", u'Q', 2, 2), ("f3", u'R', 2, 2), ("f3", u'S', 2, 2),
              ("f3", u'X', 3, 3), ("f3", u'Y', 3, 3), ("f3", u'Z', 2, 2)]
    target = sorted(target)
    
    stored = [{"f1": "A B C"}, {"f1": "D E F"}, {"f1": "A E C"},
              {"f1": "A A A"}, {"f1": "A B"}]
    
    def t(ix):
        r = ix.reader()
        assert_equal(list(r.all_stored_fields()), stored)
        assert_equal(sorted(r), target)
    
    ix = _one_segment_index()
    assert_equal(len(ix._segments()), 1)
    t(ix)
    
    ix = _multi_segment_index()
    assert_equal(len(ix._segments()), 3)
    t(ix)

def test_term_inspection():
    schema = fields.Schema(title=fields.TEXT(stored=True),
                           content=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(schema)
    writer = ix.writer()
    writer.add_document(title=u"My document",
                        content=u"AA AA BB BB CC AA AA AA BB BB CC DD EE EE")
    writer.add_document(title=u"My other document",
                        content=u"AA AB BB CC EE EE AX AX DD")
    writer.commit()
    
    reader = ix.reader()
    assert_equal(list(reader.lexicon("content")), [u'aa', u'ab', u'ax', u'bb', u'cc', u'dd', u'ee'])
    assert_equal(list(reader.expand_prefix("content", "a")), [u'aa', u'ab', u'ax'])
    assert (set(reader.all_terms())
            == set([('content', u'aa'), ('content', u'ab'), ('content', u'ax'),
                    ('content', u'bb'), ('content', u'cc'), ('content', u'dd'),
                    ('content', u'ee'), ('title', u'document'), ('title', u'my'),
                    ('title', u'other')]))
    # (text, doc_freq, index_freq)
    assert (list(reader.iter_field("content"))
            == [(u'aa', 2, 6), (u'ab', 1, 1), (u'ax', 1, 2), (u'bb', 2, 5),
                (u'cc', 2, 3), (u'dd', 2, 2), (u'ee', 2, 4)])
    assert (list(reader.iter_field("content", prefix="c"))
             == [(u'cc', 2, 3), (u'dd', 2, 2), (u'ee', 2, 4)])
    assert (list(reader.most_frequent_terms("content"))
            == [(6, u'aa'), (5, u'bb'), (4, u'ee'), (3, u'cc'), (2, u'dd')])
    assert (list(reader.most_frequent_terms("content", prefix="a"))
            == [(6, u'aa'), (2, u'ax'), (1, u'ab')])

def test_vector_postings():
    s = fields.Schema(id=fields.ID(stored=True, unique=True),
                      content=fields.TEXT(vector=fields.Positions(analyzer=analysis.StandardAnalyzer())))
    st = RamStorage()
    ix = st.create_index(s)
    
    writer = ix.writer()
    writer.add_document(id=u'1', content=u'the quick brown fox jumped over the lazy dogs')
    writer.commit()
    r = ix.reader()
    
    terms = list(r.vector_as("weight", 0, "content"))
    assert_equal(terms, [(u'brown', 1.0), (u'dogs', 1.0), (u'fox', 1.0),
                         (u'jumped', 1.0), (u'lazy', 1.0), (u'over', 1.0),
                         (u'quick', 1.0)])
    
def test_stored_fields():
    s = fields.Schema(a=fields.ID(stored=True), b=fields.STORED,
                      c=fields.KEYWORD, d=fields.TEXT(stored=True))
    st = RamStorage()
    ix = st.create_index(s)
    
    writer = ix.writer()
    writer.add_document(a=u"1", b="a", c=u"zulu", d=u"Alfa")
    writer.add_document(a=u"2", b="b", c=u"yankee", d=u"Bravo")
    writer.add_document(a=u"3", b="c", c=u"xray", d=u"Charlie")
    writer.commit()
    
    with ix.searcher() as sr:
        assert_equal(sr.stored_fields(0), {"a": u"1", "b": "a", "d": u"Alfa"})
        assert_equal(sr.stored_fields(2), {"a": u"3", "b": "c", "d": u"Charlie"})
        
        assert_equal(sr.document(a=u"1"), {"a": u"1", "b": "a", "d": u"Alfa"})
        assert_equal(sr.document(a=u"2"), {"a": u"2", "b": "b", "d": u"Bravo"})

def test_stored_fields2():
    schema = fields.Schema(content=fields.TEXT(stored=True),
                           title=fields.TEXT(stored=True),
                           summary=fields.STORED,
                           path=fields.ID(stored=True),
                           helpid=fields.KEYWORD,
                           parent=fields.KEYWORD,
                           context=fields.KEYWORD(stored=True),
                           type=fields.KEYWORD(stored=True),
                           status=fields.KEYWORD(stored=True),
                           superclass=fields.KEYWORD(stored=True),
                           exampleFor=fields.KEYWORD(stored=True),
                           chapter=fields.KEYWORD(stored=True),
                           replaces=fields.KEYWORD,
                           time=fields.STORED,
                           methods=fields.STORED,
                           exampleFile=fields.STORED,
                           )
    
    storedkeys = ["chapter", "content", "context", "exampleFile",
                  "exampleFor", "methods", "path", "status", "summary",
                  "superclass", "time", "title", "type"]
    assert_equal(storedkeys, schema.stored_names())
    
    st = RamStorage()
    ix = st.create_index(schema)
    
    writer = ix.writer()
    writer.add_document(content=u"Content of this document.",
                        title=u"This is the title",
                        summary=u"This is the summary", path=u"/main")
    writer.add_document(content=u"Second document.", title=u"Second title",
                        summary=u"Summary numero due", path=u"/second")
    writer.add_document(content=u"Third document.", title=u"Title 3",
                        summary=u"Summary treo", path=u"/san")
    writer.commit()
    ix.close()
    
    ix = st.open_index()
    with ix.searcher() as s:
        doc = s.document(path="/main")
        assert ([doc[k] for k in sorted(doc.keys())]
                == ["Content of this document.", "/main",
                    "This is the summary", "This is the title"])
    
    ix.close()
    
def test_first_id():
    schema = fields.Schema(path=fields.ID(stored=True))
    ix = RamStorage().create_index(schema)
    
    w = ix.writer()
    w.add_document(path=u"/a")
    w.add_document(path=u"/b")
    w.add_document(path=u"/c")
    w.commit()
    
    r = ix.reader()
    docid = r.first_id("path", u"/b")
    assert_equal(r.stored_fields(docid), {"path": "/b"})
    
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(path=u"/a")
    w.add_document(path=u"/b")
    w.add_document(path=u"/c")
    w.commit(merge=False)
    
    w = ix.writer()
    w.add_document(path=u"/d")
    w.add_document(path=u"/e")
    w.add_document(path=u"/f")
    w.commit(merge=False)

    w = ix.writer()
    w.add_document(path=u"/g")
    w.add_document(path=u"/h")
    w.add_document(path=u"/i")
    w.commit(merge=False)

    r = ix.reader()
    assert_equal(r.__class__, reading.MultiReader)
    docid = r.first_id("path", u"/e")
    assert_equal(r.stored_fields(docid), {"path": "/e"})











from __future__ import with_statement
import random

from nose.tools import assert_equal

from whoosh import fields, query
from whoosh.filedb.filestore import RamStorage
from whoosh.support.testing import skip_if_unavailable, TempIndex


try:
    import multiprocessing
except ImportError:
    pass
else:
    class MPFCTask(multiprocessing.Process):
        def __init__(self, storage, indexname):
            multiprocessing.Process.__init__(self)
            self.storage = storage
            self.indexname = indexname
            
        def run(self):
            ix = self.storage.open_index(self.indexname)
            with ix.searcher() as s:
                r = s.search(query.Every(), sortedby="key", limit=None)
                result = "".join([h["key"] for h in r])
                assert_equal(result, "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")


docs = ({"id": u"zulu", "num": 100, "tag": u"one", "frac": 0.75},
        {"id": u"xray", "num": -5, "tag": u"three", "frac": 2.0},
        {"id": u"yankee", "num": 3, "tag": u"two", "frac": 5.5},
        
        {"id": u"alfa", "num": 7, "tag": u"three", "frac": 2.25},
        {"id": u"tango", "num": 2, "tag": u"two", "frac": 1.75},
        {"id": u"foxtrot", "num": -800, "tag": u"two", "frac": 3.25},
        
        {"id": u"sierra", "num": 1, "tag": u"one", "frac": 4.75},
        {"id": u"whiskey", "num": 0, "tag": u"three", "frac": 5.25},
        {"id": u"bravo", "num": 582045, "tag": u"three", "frac": 1.25},
        )

def get_schema():
    return fields.Schema(id=fields.ID(stored=True),
                         num=fields.NUMERIC(stored=True),
                         frac=fields.NUMERIC(type=float, stored=True),
                         tag=fields.ID(stored=True),
                         ev=fields.ID,
                         )

def make_single_index(ix):
    w = ix.writer()
    for doc in docs:
        w.add_document(ev=u"a", **doc)
    w.commit()

def make_multi_index(ix):
    for i in xrange(0, len(docs), 3):
        w = ix.writer()
        for doc in docs[i:i+3]:
            w.add_document(ev=u"a", **doc)
        w.commit(merge=False)

def try_sort(sortedby, key, q=None, limit=None, reverse=False):
    if q is None: q = query.Term("ev", u"a")
    
    correct = [d["id"] for d in sorted(docs, key=key, reverse=reverse)][:limit]
    
    for fn in (make_single_index, make_multi_index):
        with TempIndex(get_schema()) as ix:
            fn(ix)
            with ix.searcher() as s:
                r = s.search(q, sortedby=sortedby, limit=limit, reverse=reverse)
                rids = [d["id"] for d in r]
                assert_equal(rids, correct)


def test_cached_lexicon():
    schema = fields.Schema(tag=fields.ID)
    with TempIndex(schema, "cachedlexicon") as ix:
        w = ix.writer()
        w.add_document(tag=u"sierra")
        w.add_document(tag=u"alfa")
        w.add_document(tag=u"juliet")
        w.add_document(tag=u"romeo")
        w.commit()
        
        with ix.reader() as r:
            fc = r.fieldcache("tag")
            assert_equal(list(r.lexicon("tag")), ["alfa", "juliet", "romeo", "sierra"])

# 

def test_float_cache():
    schema = fields.Schema(id=fields.STORED, num=fields.NUMERIC(type=float))
    with TempIndex(schema, "floatcache") as ix:
        w = ix.writer()
        w.add_document(id=1, num=1.5)
        w.add_document(id=2, num=-8.25)
        w.add_document(id=3, num=0.75)
        w.commit()
        
        with ix.reader() as r:
            r.fieldcache("num")
            r.unload_fieldcache("num")
            
            fc = r.fieldcache("num")
            assert not fc.hastexts
            assert_equal(fc.texts, None)
            assert_equal(fc.typecode, "f")
            assert_equal(list(fc.order), [1.5, -8.25, 0.75])

def test_long_cache():
    schema = fields.Schema(id=fields.STORED, num=fields.NUMERIC(type=long))
    with TempIndex(schema, "longcache") as ix:
        w = ix.writer()
        w.add_document(id=1, num=2858205080241)
        w.add_document(id=2, num=-3572050858202)
        w.add_document(id=3, num=4985020582043)
        w.commit()
        
        with ix.reader() as r:
            r.fieldcache("num")
            r.unload_fieldcache("num")
            
            fc = r.fieldcache("num")
            assert not fc.hastexts
            assert_equal(fc.texts, None)
            assert_equal(fc.typecode, "q")
            assert_equal(list(fc.order), [2858205080241, -3572050858202, 4985020582043])

def test_shared_cache():
    with TempIndex(get_schema(), "sharedcache") as ix:
        make_single_index(ix)
        r1 = ix.reader()
        fc1 = r1.fieldcache("id")
        
        r2 = ix.reader()
        fc2 = r2.fieldcache("id")
        
        assert fc1 is fc2
        
        r3 = ix.reader()
        assert r3.fieldcache_loaded("id")
        
        r1.close()
        r2.close()
        del r1, fc1, r2, fc2
        import gc
        gc.collect()
        
        assert not r3.fieldcache_loaded("id")
        r3.close()
        
def test_sortedby():
    try_sort("id", lambda d: d["id"])
    try_sort("id", lambda d: d["id"], limit=5)
    try_sort("id", lambda d: d["id"], reverse=True)
    try_sort("id",  lambda d: d["id"], limit=5, reverse=True)

def test_multisort():
    try_sort(("tag", "id"), lambda d: (d["tag"], d["id"]))
    try_sort(("tag", "id"), lambda d: (d["tag"], d["id"]), reverse=True)
    try_sort(("tag", "id"), lambda d: (d["tag"], d["id"]), limit=5)
    try_sort(("tag", "id"), lambda d: (d["tag"], d["id"]), reverse=True, limit=5)

def test_numeric():
    try_sort("num", lambda d: d["num"])
    try_sort("num", lambda d: d["num"], reverse=True)
    try_sort("num", lambda d: d["num"], limit=5)
    try_sort("frac", lambda d: d["frac"])

def test_empty_field():
    schema = fields.Schema(id=fields.STORED, key=fields.KEYWORD)
    with TempIndex(schema, "emptysort") as ix:
        w = ix.writer()
        w.add_document(id=1)
        w.add_document(id=2)
        w.add_document(id=3)
        w.commit()
        
        with ix.searcher() as s:
            r = s.search(query.Every(), sortedby="key")
            assert_equal([h["id"] for h in r], [1, 2, 3])

def test_page_sorted():
    schema = fields.Schema(key=fields.ID(stored=True))
    with TempIndex(schema, "pagesorted") as ix:
        domain = list(u"abcdefghijklmnopqrstuvwxyz")
        random.shuffle(domain)
        
        w = ix.writer()
        for char in domain:
            w.add_document(key=char)
        w.commit()
        
        with ix.searcher() as s:
            rp = s.search_page(query.Every(), 1, pagelen=5, sortedby="key")
            assert_equal("".join([h["key"] for h in rp]), "abcde")
            assert_equal(rp[10:], [])
            
            rp = s.search_page(query.Term("key", "glonk"), 1, pagelen=5, sortedby="key")
            assert_equal(len(rp), 0)
            assert rp.is_last_page()

@skip_if_unavailable("multiprocessing")
def test_mp_fieldcache():
    schema = fields.Schema(key=fields.KEYWORD(stored=True))
    with TempIndex(schema, "mpfieldcache") as ix:
        domain = list(u"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
        random.shuffle(domain)
        w = ix.writer()
        for char in domain:
            w.add_document(key=char)
        w.commit()
        
        tasks = [MPFCTask(ix.storage, ix.indexname) for _ in xrange(4)]
        for task in tasks:
            task.start()
        for task in tasks:
            task.join()

def test_field_facets():
    def check(method):
        with TempIndex(get_schema()) as ix:
            method(ix)
            with ix.searcher() as s:
                q = query.Every()
                groups = s.categorize_query(q, "tag")
                assert (sorted(groups.items())
                        == [(u'one', [0L, 6L]), (u'three', [1L, 3L, 7L, 8L]),
                            (u'two', [2L, 4L, 5L])])
    
    check(make_single_index)
    check(make_multi_index)

def test_query_facets():
    schema = fields.Schema(value=fields.ID(stored=True))
    with TempIndex(schema, "queryfacets") as ix:
        w = ix.writer()
        alphabet = list(u"abcdefghijklmnopqrstuvwxyz")
        random.shuffle(alphabet)
        
        for letter in alphabet:
            w.add_document(value=letter)
        w.commit()
        
        with ix.searcher() as s:
            q1 = query.TermRange("value", u"a", u"i")
            q2 = query.TermRange("value", u"j", u"r")
            q3 = query.TermRange("value", u"s", u"z")
            s.define_facets("range", {"a-i": q1, "j-r": q2, "s-z": q3},
                            save=False)
            
            def check(groups):
                for key in groups.keys():
                    groups[key] = "".join(sorted([s.stored_fields(id)["value"]
                                                  for id in groups[key]]))
                assert_equal(groups, {'a-i': u'abcdefghi', 'j-r': u'jklmnopqr',
                                      's-z': u'stuvwxyz'})
            
            check(s.categorize_query(query.Every(), "range"))

            r = s.search(query.Every(), groupedby="range")
            check(r.groups("range"))
            
        with ix.searcher() as s:
            assert not s.reader().fieldcache_available("range")

def test_multifacet():
    schema = fields.Schema(tag=fields.ID(stored=True),
                           size=fields.ID(stored=True))
    with TempIndex(schema, "multifacet") as ix:
        w = ix.writer()
        w.add_document(tag=u"alfa", size=u"small")
        w.add_document(tag=u"bravo", size=u"medium")
        w.add_document(tag=u"alfa", size=u"large")
        w.add_document(tag=u"bravo", size=u"small")
        w.add_document(tag=u"alfa", size=u"medium")
        w.add_document(tag=u"bravo", size=u"medium")
        w.commit()
        
        correct = {(u'bravo', u'medium'): [1, 5], (u'alfa', u'large'): [2],
                   (u'alfa', u'medium'): [4], (u'alfa', u'small'): [0],
                   (u'bravo', u'small'): [3]}
        
        with ix.searcher() as s:
            cats = s.categorize_query(query.Every(), ("tag", "size"))
            assert_equal(cats, correct)
            
            r = s.search(query.Every(), groupedby=[("tag", "size")])
            cats = r.groups(("tag", "size"))
            assert_equal(cats, correct)

def test_sort_filter():
    schema = fields.Schema(group=fields.ID(stored=True), key=fields.ID(stored=True))
    groups = u"alfa bravo charlie".split()
    keys = u"abcdefghijklmnopqrstuvwxyz"
    source = []
    for i in xrange(100):
        key = keys[i % len(keys)]
        group = groups[i % len(groups)]
        source.append({"key": key, "group": group})
    source.sort(key=lambda x: (x["key"], x["group"]))
    
    sample = source[:]
    random.shuffle(sample)
    
    with TempIndex(schema, "sortfilter") as ix:
        w = ix.writer()
        for i, fs in enumerate(sample):
            w.add_document(**fs)
            i += 1
            if not i % 26:
                w.commit(merge=False)
                w = ix.writer()
        w.commit()
        
        fq = query.Term("group", u"bravo")
        
        with ix.searcher() as s:
            r = s.search(query.Every(), sortedby=("key", "group"), filter=fq, limit=20)
            assert_equal([h.fields() for h in r],
                         [d for d in source if d["group"] == "bravo"][:20])
            
            fq = query.Term("group", u"bravo")
            r = s.search(query.Every(), sortedby=("key", "group"), filter=fq, limit=None)
            assert_equal([h.fields() for h in r],
                         [d for d in source if d["group"] == "bravo"])
            
        ix.optimize()
        
        with ix.searcher() as s:
            r = s.search(query.Every(), sortedby=("key", "group"), filter=fq, limit=20)
            assert_equal([h.fields() for h in r],
                         [d for d in source if d["group"] == "bravo"][:20])
            
            fq = query.Term("group", u"bravo")
            r = s.search(query.Every(), sortedby=("key", "group"), filter=fq, limit=None)
            assert_equal([h.fields() for h in r],
                         [d for d in source if d["group"] == "bravo"])

def test_custom_sort():
    schema = fields.Schema(name=fields.ID(stored=True),
                           price=fields.NUMERIC,
                           quant=fields.NUMERIC)
    
    with TempIndex(schema, "customsort") as ix:
        w = ix.writer()
        w.add_document(name=u"A", price=200, quant=9)
        w.add_document(name=u"E", price=300, quant=4)
        w.add_document(name=u"F", price=200, quant=8)
        w.add_document(name=u"D", price=150, quant=5)
        w.add_document(name=u"B", price=250, quant=11)
        w.add_document(name=u"C", price=200, quant=10)
        w.commit()
        
        with ix.searcher() as s:
            cs = s.sorter()
            cs.add_field("price")
            cs.add_field("quant", reverse=True)
            print "crit=", cs.criteria
            print "is_simple=", cs.is_simple()
            r = cs.sort_query(query.Every(), limit=None)
            assert_equal([hit["name"] for hit in r], list(u"DCAFBE"))
            
def test_sorting_function():
    schema = fields.Schema(id=fields.STORED, text=fields.TEXT(stored=True, vector=True))
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    domain = ("alfa", "bravo", "charlie")
    count = 1
    for w1 in domain:
        for w2 in domain:
            for w3 in domain:
                for w4 in domain:
                    w.add_document(id=count, text=u" ".join((w1, w2, w3, w4)))
                    count += 1
    w.commit()
    
    def fn(searcher, docnum):
        v = dict(searcher.vector_as("frequency", docnum, "text"))
        # Give high score to documents that have equal number of "alfa"
        # and "bravo"
        return 1.0 / (abs(v.get("alfa", 0) - v.get("bravo", 0)) + 1.0)
    
    with ix.searcher() as s:
        q = query.And([query.Term("text", u"alfa"), query.Term("text", u"bravo")])
        
        r = [hit["text"] for hit in s.sort_query_using(q, fn)]
        for t in r[:10]:
            tks = t.split()
            assert_equal(tks.count("alfa"), tks.count("bravo"))

def test_natural_order():
    schema = fields.Schema(id=fields.STORED, tag=fields.ID)
    
    # Single segment
    
    ix = RamStorage().create_index(schema)
    domain = u"one two three four five six seven eight nine".split()
    w = ix.writer()
    for word in domain:
        w.add_document(id=word, tag=u"tag")
    w.commit(merge=False)
    
    with ix.searcher() as s:
        r = s.search(query.Term("tag", u"tag"), scored=False)
        assert_equal([hit["id"] for hit in r], domain)
        
        r = s.search(query.Term("tag", u"tag"), scored=False, reverse=True)
        assert_equal([hit["id"] for hit in r], list(reversed(domain)))
        
        r = s.search(query.Term("tag", u"tag"), scored=False, limit=3)
        assert_equal([hit["id"] for hit in r], u"one two three".split())
        
        r = s.search(query.Term("tag", u"tag"), scored=False, reverse=True, limit=3)
        assert_equal([hit["id"] for hit in r], u"nine eight seven".split())
    
    # Multiple segments
    
    ix = RamStorage().create_index(schema)
    domain = u"one two three four five six seven eight nine".split()
    for i in xrange(0, len(domain), 3):
        w = ix.writer()
        for j in xrange(3):
            w.add_document(id=domain[i+j], tag=u"tag")
        w.commit(merge=False)
    
    with ix.searcher() as s:
        r = s.search(query.Term("tag", u"tag"), scored=False)
        assert_equal([hit["id"] for hit in r], domain)
        
        r = s.search(query.Term("tag", u"tag"), scored=False, reverse=True)
        assert_equal([hit["id"] for hit in r], list(reversed(domain)))
        
        r = s.search(query.Term("tag", u"tag"), scored=False, limit=3)
        assert_equal([hit["id"] for hit in r], u"one two three".split())
        
        r = s.search(query.Term("tag", u"tag"), scored=False, reverse=True, limit=3)
        assert_equal([hit["id"] for hit in r], u"nine eight seven".split())





#def test_custom_sort2():
#    from array import array
#    from whoosh.searching import Results
#    
#    class CustomSort(object):
#        def __init__(self, *criteria):
#            self.criteria = criteria
#            self.arrays = None
#            
#        def cache(self, searcher):
#            self.arrays = []
#            r = searcher.reader()
#            for name, reverse in self.criteria:
#                arry = array("i", [0] * r.doc_count_all())
#                field = ix.schema[name]
#                for i, (token, _) in enumerate(field.sortable_values(r, name)):
#                    if reverse: i = 0 - i
#                    postings = r.postings(name, token)
#                    for docid in postings.all_ids():
#                        arry[docid] = i
#                self.arrays.append(arry)
#                
#        def key_fn(self, docnum):
#            return tuple(arry[docnum] for arry in self.arrays)
#        
#        def sort_query(self, searcher, q):
#            if self.arrays is None:
#                self.cache(searcher)
#            
#            return self._results(searcher, q, searcher.docs_for_query(q))
#        
#        def sort_all(self, searcher):
#            if self.arrays is None:
#                self.cache(searcher)
#            
#            return self._results(searcher, None, searcher.reader().all_doc_ids())
#            
#        def _results(self, searcher, q, docnums):
#            docnums = sorted(docnums, key=self.key_fn)
#            return Results(searcher, q, [(None, docnum) for docnum in docnums], None)
#            
#    
#    schema = fields.Schema(name=fields.ID(stored=True),
#                           price=fields.NUMERIC,
#                           quant=fields.NUMERIC)
#    
#    with TempIndex(schema, "customsort") as ix:
#        w = ix.writer()
#        w.add_document(name=u"A", price=200, quant=9)
#        w.add_document(name=u"E", price=300, quant=4)
#        w.add_document(name=u"F", price=200, quant=8)
#        w.add_document(name=u"D", price=150, quant=5)
#        w.add_document(name=u"B", price=250, quant=11)
#        w.add_document(name=u"C", price=200, quant=10)
#        w.commit()
#        
#        cs = CustomSort(("price", False), ("quant", True))
#        with ix.searcher() as s:
#            assert_equal([hit["name"] for hit in cs.sort_query(s, query.Every())],
#                          list("DCAFBE"))
                    
        
        









from __future__ import with_statement

from nose.tools import assert_equal

from whoosh import fields
from whoosh.support.testing import TempIndex


def test_addfield():
    schema = fields.Schema(id=fields.ID(stored=True), content=fields.TEXT)
    with TempIndex(schema, "addfield") as ix:
        w = ix.writer()
        w.add_document(id=u"a", content=u"alfa")
        w.add_document(id=u"b", content=u"bravo")
        w.add_document(id=u"c", content=u"charlie")
        w.commit()
        
        ix.add_field("added", fields.KEYWORD(stored=True))
        
        w = ix.writer()
        w.add_document(id=u"d", content=u"delta", added=u"fourth")
        w.add_document(id=u"e", content=u"echo", added=u"fifth")
        w.commit(merge=False)
        
        with ix.searcher() as s:
            assert_equal(s.document(id=u"d"), {"id": "d", "added": "fourth"})
            assert_equal(s.document(id=u"b"), {"id": "b"})

def test_removefield():
    schema = fields.Schema(id=fields.ID(stored=True),
                           content=fields.TEXT,
                           city=fields.KEYWORD(stored=True))
    with TempIndex(schema, "removefield") as ix:
        w = ix.writer()
        w.add_document(id=u"b", content=u"bravo", city=u"baghdad")
        w.add_document(id=u"c", content=u"charlie", city=u"cairo")
        w.add_document(id=u"d", content=u"delta", city=u"dakar")
        w.commit()
        
        with ix.searcher() as s:
            assert_equal(s.document(id=u"c"), {"id": "c", "city": "cairo"})
        
        w = ix.writer()
        w.remove_field("content")
        w.remove_field("city")
        w.commit()

        ixschema = ix._current_schema()
        assert_equal(ixschema.names(), ["id"])
        assert_equal(ixschema.stored_names(), ["id"])
        
        with ix.searcher() as s:
            assert ("content", u"charlie") not in s.reader()
            assert_equal(s.document(id=u"c"), {"id": u"c"})

def test_optimize_away():
    schema = fields.Schema(id=fields.ID(stored=True),
                           content=fields.TEXT,
                           city=fields.KEYWORD(stored=True))
    with TempIndex(schema, "optimizeaway") as ix:
        w = ix.writer()
        w.add_document(id=u"b", content=u"bravo", city=u"baghdad")
        w.add_document(id=u"c", content=u"charlie", city=u"cairo")
        w.add_document(id=u"d", content=u"delta", city=u"dakar")
        w.commit()
        
        with ix.searcher() as s:
            assert_equal(s.document(id=u"c"), {"id": "c", "city": "cairo"})
        
        w = ix.writer()
        w.remove_field("content")
        w.remove_field("city")
        w.commit(optimize=True)
        
        with ix.searcher() as s:
            assert ("content", u"charlie") not in s.reader()
            assert_equal(s.document(id=u"c"), {"id": u"c"})
        






# Copyright 2010 Matt Chaput. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY MATT CHAPUT ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL MATT CHAPUT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of Matt Chaput.

import re
from datetime import datetime, timedelta

from whoosh.qparser import BasicSyntax, ErrorToken, Plugin, RangePlugin, Group, Word
from whoosh.support.relativedelta import relativedelta
from whoosh.support.times import (adatetime, timespan, fill_in, is_void,
                                  TimeError, relative_days)


class DateParseError(Exception):
    "Represents an error in parsing date text."


# Utility functions

def rcompile(pattern):
    """Just a shortcut to call re.compile with a standard set of flags.
    """
    
    return re.compile(pattern, re.IGNORECASE | re.UNICODE)


def print_debug(level, msg, *args):
    if level > 0:
        print ("  " * (level - 1)) + (msg % args)


# Parser element objects

class Props(object):
    """A dumb little object that just puts copies a dictionary into attibutes
    so I can use dot syntax instead of square bracket string item lookup and
    save a little bit of typing. Used by :class:`Regex`.
    """
    
    def __init__(self, **args):
        self.__dict__ = args
    
    def __repr__(self):
        return repr(self.__dict__)
    
    def get(self, key, default=None):
        return self.__dict__.get(key, default)


class ParserBase(object):
    """Base class for date parser elements.
    """
    
    def to_parser(self, e):
        if isinstance(e, basestring):
            return Regex(e)
        else:
            return e
    
    def parse(self, text, dt, pos=0, debug=-9999):
        raise NotImplementedError
    
    def date_from(self, text, dt=None, pos=0, debug=-9999):
        if dt is None:
            dt = datetime.now()
        
        d, pos = self.parse(text, dt, pos, debug + 1)
        return d


class MultiBase(ParserBase):
    """Base class for date parser elements such as Sequence and Bag that
    have sub-elements.
    """
    
    def __init__(self, elements, name=None):
        """
        :param elements: the sub-elements to match.
        :param name: a name for this element (for debugging purposes only).
        """
        
        self.elements = [self.to_parser(e) for e in elements]
        self.name = name
        
    def __repr__(self):
        return "%s<%s>%r" % (self.__class__.__name__, self.name or '', self.elements)


class Sequence(MultiBase):
    """Merges the dates parsed by a sequence of sub-elements.
    """
    
    def __init__(self, elements, sep="(\\s+|\\s*,\\s*)", name=None,
                 progressive=False):
        """
        :param elements: the sequence of sub-elements to parse.
        :param sep: a separator regular expression to match between elements,
            or None to not have separators.
        :param name: a name for this element (for debugging purposes only).
        :param progressive: if True, elements after the first do not need to
            match. That is, for elements (a, b, c) and progressive=True, the
            sequence matches like ``a[b[c]]``.
        """
        
        super(Sequence, self).__init__(elements, name)
        self.sep_pattern = sep
        if sep:
            self.sep_expr = rcompile(sep)
        else:
            self.sep_expr = None
        self.progressive = progressive
    
    def parse(self, text, dt, pos=0, debug=-9999):
        d = adatetime()
        first = True
        foundall = False
        failed = False
        
        print_debug(debug, "Seq %s sep=%r text=%r", self.name, self.sep_pattern, text[pos:])
        for e in self.elements:
            print_debug(debug, "Seq %s text=%r", self.name, text[pos:])
            if self.sep_expr and not first:
                print_debug(debug, "Seq %s looking for sep", self.name)
                m = self.sep_expr.match(text, pos)
                if m:
                    pos = m.end()
                else:
                    print_debug(debug, "Seq %s didn't find sep", self.name)
                    break
            
            print_debug(debug, "Seq %s trying=%r at=%s", self.name, e, pos)
            
            try:
                at, newpos = e.parse(text, dt, pos=pos, debug=debug + 1)
            except TimeError:
                failed = True
                break
            
            print_debug(debug, "Seq %s result=%r", self.name, at)
            if not at:
                break
            pos = newpos
            
            print_debug(debug, "Seq %s adding=%r to=%r", self.name, at, d)
            try:
                d = fill_in(d, at)
            except TimeError:
                print_debug(debug, "Seq %s Error in fill_in", self.name)
                failed = True
                break
            print_debug(debug, "Seq %s filled date=%r", self.name, d)
            
            first = False
        else:
            foundall = True
        
        if not failed and (foundall or (not first and self.progressive)):
            print_debug(debug, "Seq %s final=%r", self.name, d)
            return (d, pos)
        else:
            print_debug(debug, "Seq %s failed", self.name)
            return (None, None)


class Combo(Sequence):
    """Parses a sequence of elements in order and combines the dates parsed
    by the sub-elements somehow. The default behavior is to accept two dates
    from the sub-elements and turn them into a range. 
    """
    
    def __init__(self, elements, fn=None, sep="(\\s+|\\s*,\\s*)", min=2, max=2,
                 name=None):
        """
        :param elements: the sequence of sub-elements to parse.
        :param fn: a function to run on all dates found. It should return a
            datetime, adatetime, or timespan object. If this argument is None,
            the default behavior accepts two dates and returns a timespan.
        :param sep: a separator regular expression to match between elements,
            or None to not have separators.
        :param min: the minimum number of dates required from the sub-elements.
        :param max: the maximum number of dates allowed from the sub-elements.
        :param name: a name for this element (for debugging purposes only).
        """
        
        super(Combo, self).__init__(elements, sep=sep, name=name)
        self.fn = fn
        self.min = min
        self.max = max
    
    def parse(self, text, dt, pos=0, debug=-9999):
        dates = []
        first = True
        
        print_debug(debug, "Combo %s sep=%r text=%r", self.name, self.sep_pattern, text[pos:])
        for e in self.elements:
            if self.sep_expr and not first:
                print_debug(debug, "Combo %s looking for sep at %r", self.name, text[pos:])
                m = self.sep_expr.match(text, pos)
                if m:
                    pos = m.end()
                else:
                    print_debug(debug, "Combo %s didn't find sep", self.name)
                    return (None, None)
            
            print_debug(debug, "Combo %s trying=%r", self.name, e)
            try:
                at, pos = e.parse(text, dt, pos, debug + 1)
            except TimeError:
                at, pos = None, None
            
            print_debug(debug, "Combo %s result=%r", self.name, at)
            if at is None:
                return (None, None)
            
            first = False
            if is_void(at):
                continue
            if len(dates) == self.max:
                print_debug(debug, "Combo %s length > %s", self.name, self.max)
                return (None, None)
            dates.append(at)
        
        print_debug(debug, "Combo %s dates=%r", self.name, dates)
        if len(dates) < self.min:
            print_debug(debug, "Combo %s length < %s", self.name, self.min)
            return (None, None)
        
        return (self.dates_to_timespan(dates), pos)
    
    def dates_to_timespan(self, dates):
        if self.fn:
            return self.fn(dates)
        elif len(dates) == 2:
            return timespan(dates[0], dates[1])
        else:
            raise DateParseError("Don't know what to do with %r" % (dates, ))


class Choice(MultiBase):
    """Returns the date from the first of its sub-elements that matches.
    """
    
    def parse(self, text, dt, pos=0, debug=-9999):
        print_debug(debug, "Choice %s text=%r", self.name, text[pos:])
        for e in self.elements:
            print_debug(debug, "Choice %s trying=%r", self.name, e)
            
            try:
                d, newpos = e.parse(text, dt, pos, debug + 1)
            except TimeError:
                d, newpos = None, None
            if d:
                print_debug(debug, "Choice %s matched", self.name)
                return (d, newpos)
        print_debug(debug, "Choice %s no match", self.name)
        return (None, None)


class Bag(MultiBase):
    """Parses its sub-elements in any order and merges the dates.
    """
    
    def __init__(self, elements, sep="(\\s+|\\s*,\\s*)", onceper=True,
                 requireall=False, allof=None, anyof=None, name=None):
        """
        :param elements: the sub-elements to parse.
        :param sep: a separator regular expression to match between elements,
            or None to not have separators.
        :param onceper: only allow each element to match once.
        :param requireall: if True, the sub-elements can match in any order,
            but they must all match.
        :param allof: a list of indexes into the list of elements. When this
            argument is not None, this element matches only if all the
            indicated sub-elements match.
        :param allof: a list of indexes into the list of elements. When this
            argument is not None, this element matches only if any of the
            indicated sub-elements match.
        :param name: a name for this element (for debugging purposes only).
        """
        
        super(Bag, self).__init__(elements, name)
        self.sep_expr = rcompile(sep)
        self.onceper = onceper
        self.requireall = requireall
        self.allof = allof
        self.anyof = anyof
    
    def parse(self, text, dt, pos=0, debug=-9999):
        first = True
        d = adatetime()
        seen = [False] * len(self.elements)
        
        while True:
            newpos = pos
            print_debug(debug, "Bag %s text=%r", self.name, text[pos:])
            if not first:
                print_debug(debug, "Bag %s looking for sep", self.name)
                m = self.sep_expr.match(text, pos)
                if m:
                    newpos = m.end()
                else:
                    print_debug(debug, "Bag %s didn't find sep", self.name)
                    break
            
            for i, e in enumerate(self.elements):
                print_debug(debug, "Bag %s trying=%r", self.name, e)
                
                try:
                    at, xpos = e.parse(text, dt, newpos, debug + 1)
                except TimeError:
                    at, xpos = None, None
                    
                print_debug(debug, "Bag %s result=%r", self.name, at)
                if at:
                    if self.onceper and seen[i]:
                        return (None, None)
                    
                    d = fill_in(d, at)
                    newpos = xpos
                    seen[i] = True
                    break
            else:
                break
            
            pos = newpos
            if self.onceper and all(seen):
                break
            
            first = False
        
        if (not any(seen)
            or (self.allof and not all(seen[pos] for pos in self.allof))
            or (self.anyof and not any(seen[pos] for pos in self.anyof))
            or (self.requireall and not all(seen))):
            return (None, None)
        
        print_debug(debug, "Bag %s final=%r", self.name, d)
        return (d, pos)
    

class Optional(ParserBase):
    """Wraps a sub-element to indicate that the sub-element is optional.
    """
    
    def __init__(self, element):
        self.element = self.to_parser(element)
    
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.element)
    
    def parse(self, text, dt, pos=0, debug=-9999):
        try:
            d, pos = self.element.parse(text, dt, pos, debug + 1)
        except TimeError:
            d, pos = None, None
            
        if d:
            return (d, pos)
        else:
            return (adatetime(), pos)


class ToEnd(ParserBase):
    """Wraps a sub-element and requires that the end of the sub-element's match
    be the end of the text.
    """
    
    def __init__(self, element):
        self.element = element
        
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.element)
    
    def parse(self, text, dt, pos=0, debug=-9999):
        try:
            d, pos = self.element.parse(text, dt, pos, debug + 1)
        except TimeError:
            d, pos = None, None
            
        if d and pos == len(text):
            return (d, pos)
        else:
            return (None, None)


class Regex(ParserBase):
    """Matches a regular expression and maps named groups in the pattern to
    datetime attributes using a function or overridden method.
    
    There are two points at which you can customize the behavior of this class,
    either by supplying functions to the initializer or overriding methods.
    
    * The ``modify`` function or ``modify_props`` method takes a ``Props``
      object containing the named groups and modifies its values (in place).
    * The ``fn`` function or ``props_to_date`` method takes a ``Props`` object
      and the base datetime and returns an adatetime/datetime.
    """
    
    fn = None
    modify = None
    
    def __init__(self, pattern, fn=None, modify=None):
        self.pattern = pattern
        self.expr = rcompile(pattern)
        self.fn = fn
        self.modify = modify
    
    def __repr__(self):
        return "<%r>" % (self.pattern, )
    
    def parse(self, text, dt, pos=0, debug=-9999):
        m = self.expr.match(text, pos)
        if not m:
            return (None, None)
        
        props = self.extract(m)
        self.modify_props(props)
        
        try:
            d = self.props_to_date(props, dt)
        except TimeError:
            d = None
        
        if d:
            return (d, m.end())
        else:
            return (None, None)
    
    def extract(self, match):
        d = match.groupdict()
        for key, value in d.iteritems():
            try:
                value = int(value)
                d[key] = value
            except (ValueError, TypeError):
                pass
        return Props(**d)
    
    def modify_props(self, props):
        if self.modify:
            self.modify(props)
            
    def props_to_date(self, props, dt):
        if self.fn:
            return self.fn(props, dt)
        else:
            args = {}
            for key in adatetime.units:
                args[key] = props.get(key)
            return adatetime(**args)

    
class Month(Regex):
    def __init__(self, *patterns):
        self.patterns = patterns
        self.exprs = [rcompile(pat) for pat in self.patterns]
        
        self.pattern = ("(?P<month>"
                        + "|".join("(%s)" % pat for pat in self.patterns)
                        + ")")
        self.expr = rcompile(self.pattern)
        
    def modify_props(self, p):
        text = p.month
        for i, expr in enumerate(self.exprs):
            m = expr.match(text)
            if m:
                p.month = i + 1
                break
            

class PlusMinus(Regex):
    def __init__(self, years, months, weeks, days, hours, minutes, seconds):
        rel_years = "((?P<years>[0-9]+) *(%s))?" % years
        rel_months = "((?P<months>[0-9]+) *(%s))?" % months
        rel_weeks = "((?P<weeks>[0-9]+) *(%s))?" % weeks
        rel_days = "((?P<days>[0-9]+) *(%s))?" % days
        rel_hours = "((?P<hours>[0-9]+) *(%s))?" % hours
        rel_mins = "((?P<mins>[0-9]+) *(%s))?" % minutes
        rel_secs = "((?P<secs>[0-9]+) *(%s))?" % seconds
        
        self.pattern = ("(?P<dir>[+-]) *%s *%s *%s *%s *%s *%s *%s(?=(\\W|$))"
                        % (rel_years, rel_months, rel_weeks, rel_days,
                           rel_hours, rel_mins, rel_secs))
        self.expr = rcompile(self.pattern)
        
    def props_to_date(self, p, dt):
        if p.dir == "-":
            dir = -1
        else:
            dir = 1
    
        delta = relativedelta(years=(p.get("years") or 0) * dir,
                              months=(p.get("months") or 0) * dir,
                              weeks=(p.get("weeks") or 0) * dir,
                              days=(p.get("days") or 0) * dir,
                              hours=(p.get("hours") or 0) * dir,
                              minutes=(p.get("mins") or 0) * dir,
                              seconds=(p.get("secs") or 0) * dir)
        return dt + delta


class Daynames(Regex):
    def __init__(self, next, last, daynames):
        self.next_pattern = next
        self.last_pattern = last
        self._dayname_exprs = tuple(rcompile(pat) for pat in daynames)
        dn_pattern = "|".join(daynames)
        self.pattern = "(?P<dir>%s|%s) +(?P<day>%s)(?=(\\W|$))" % (next, last, dn_pattern)
        self.expr = rcompile(self.pattern)
    
    def props_to_date(self, p, dt):
        if re.match(p.dir, self.last_pattern):
            dir = -1
        else:
            dir = 1
        
        for daynum, expr in enumerate(self._dayname_exprs):
            m = expr.match(p.day)
            if m:
                break
        current_daynum = dt.weekday()
        days_delta = relative_days(current_daynum, daynum, dir)
        
        d = dt.date() + timedelta(days=days_delta)
        return adatetime(year=d.year, month=d.month, day=d.day)


class Time12(Regex):
    def __init__(self):
        self.pattern = "(?P<hour>[1-9]|10|11|12)(:(?P<mins>[0-5][0-9])(:(?P<secs>[0-5][0-9])(\\.(?P<usecs>[0-9]{1,5}))?)?)?\\s*(?P<ampm>am|pm)(?=(\\W|$))"
        self.expr = rcompile(self.pattern)

    def props_to_date(self, p, dt):
        isam = p.ampm.lower().startswith("a")
        
        if p.hour == 12:
            if isam:
                hr = 0
            else:
                hr = 12
        else:
            hr = p.hour
            if not isam:
                hr += 12
        
        return adatetime(hour=hr, minute=p.mins, second=p.secs, microsecond=p.usecs)


# Top-level parser classes

class DateParser(object):
    """Base class for locale-specific parser classes.
    """
    
    day = Regex("(?P<day>([123][0-9])|[1-9])(?=(\\W|$))(?!=:)",
                lambda p, dt: adatetime(day=p.day))
    year = Regex("(?P<year>[0-9]{4})(?=(\\W|$))",
                 lambda p, dt: adatetime(year=p.year))
    time24 = Regex("(?P<hour>([0-1][0-9])|(2[0-3])):(?P<mins>[0-5][0-9])(:(?P<secs>[0-5][0-9])(\\.(?P<usecs>[0-9]{1,5}))?)?(?=(\\W|$))",
                   lambda p, dt: adatetime(hour=p.hour, minute=p.mins, second=p.secs, microsecond=p.usecs))
    time12 = Time12()
    
    def __init__(self):
        simple_year = "(?P<year>[0-9]{4})"
        simple_month = "(?P<month>[0-1][0-9])"
        simple_day = "(?P<day>[0-3][0-9])"
        simple_hour = "(?P<hour>([0-1][0-9])|(2[0-3]))"
        simple_minute = "(?P<minute>[0-5][0-9])"
        simple_second = "(?P<second>[0-5][0-9])"
        simple_usec = "(?P<microsecond>[0-9]{6})"
        
        simple_seq = Sequence((simple_year, simple_month, simple_day, simple_hour,
                               simple_minute, simple_second, simple_usec),
                               sep="[- .:/]*", name="simple", progressive=True)
        self.simple = Sequence((simple_seq, "(?=(\\s|$))"), sep='')
        
        self.setup()
    
    def setup(self):
        raise NotImplementedError
    
    #
    
    def get_parser(self):
        return self.all
    
    def parse(self, text, dt, pos=0, debug=-9999):
        parser = self.get_parser()
        
        d, newpos = parser.parse(text, dt, pos=pos, debug=debug)
        if isinstance(d, (adatetime, timespan)):
            d = d.disambiguated(dt)
        
        return (d, newpos)
    
    def date_from(self, text, basedate=None, pos=0, debug=-9999, toend=True):
        if basedate is None:
            basedate = datetime.utcnow()
        
        parser = self.get_parser()
        if toend:
            parser = ToEnd(parser)
        
        d = parser.date_from(text, basedate, pos=pos, debug=debug)
        if isinstance(d, (adatetime, timespan)):
            d = d.disambiguated(basedate)
        return d
    
    
class English(DateParser):
    day = Regex("(?P<day>([123][0-9])|[1-9])(st|nd|rd|th)?(?=(\\W|$))",
                lambda p, dt: adatetime(day=p.day))
    
    def setup(self):
        self.plusdate = PlusMinus("years|year|yrs|yr|ys|y",
                                  "months|month|mons|mon|mos|mo",
                                  "weeks|week|wks|wk|ws|w",
                                  "days|day|dys|dy|ds|d",
                                  "hours|hour|hrs|hr|hs|h",
                                  "minutes|minute|mins|min|ms|m",
                                  "seconds|second|secs|sec|s")
        
        self.dayname = Daynames("next", "last",
                                ("monday|mon|mo", "tuesday|tues|tue|tu",
                                 "wednesday|wed|we", "thursday|thur|thu|th",
                                 "friday|fri|fr", "saturday|sat|sa",
                                 "sunday|sun|su"))
        
        midnight = Regex("midnight", lambda p, dt: adatetime(hour=0, minute=0, second=0, microsecond=0))
        noon = Regex("noon", lambda p, dt: adatetime(hour=12, minute=0, second=0, microsecond=0))
        now = Regex("now", lambda p, dt: dt)
        self.time = Choice((self.time12, self.time24, midnight, noon, now), name="time")
        
        def tomorrow_to_date(p, dt):
            d = dt.date() + timedelta(days=+1)
            return adatetime(year=d.year, month=d.month, day=d.day)
        tomorrow = Regex("tomorrow", tomorrow_to_date)
        
        def yesterday_to_date(p, dt):
            d = dt.date() + timedelta(days=-1)
            return adatetime(year=d.year, month=d.month, day=d.day)
        yesterday = Regex("yesterday", yesterday_to_date)
        
        thisyear = Regex("this year", lambda p, dt: adatetime(year=dt.year))
        thismonth = Regex("this month", lambda p, dt: adatetime(year=dt.year, month=dt.month))
        today = Regex("today", lambda p, dt: adatetime(year=dt.year, month=dt.month, day=dt.day))
        
        self.month = Month("january|jan", "february|febuary|feb", "march|mar",
                           "april|apr", "may", "june|jun", "july|jul", "august|aug",
                           "september|sept|sep", "october|oct", "november|nov",
                           "december|dec")
        
        # If you specify a day number you must also specify a month... this
        # Choice captures that constraint
        
        self.dmy = Choice((Sequence((self.day, self.month, self.year), name="dmy"),
                           Sequence((self.month, self.day, self.year), name="mdy"),
                           Sequence((self.year, self.month, self.day), name="ymd"),
                           Sequence((self.year, self.day, self.month), name="ydm"),
                           Sequence((self.day, self.month), name="dm"),
                           Sequence((self.month, self.day), name="md"),
                           Sequence((self.month, self.year), name="my"),
                           self.month, self.year, self.dayname, tomorrow,
                           yesterday, thisyear, thismonth, today, now,
                           ), name="date")
        
        self.datetime = Bag((self.time, self.dmy), name="datetime")
        self.bundle = Choice((self.plusdate, self.datetime, self.simple), name="bundle")
        self.torange = Combo((self.bundle, "to", self.bundle), name="torange")
        
        self.all = Choice((self.torange, self.bundle), name="all")
        

# QueryParser plugin

class DateParserPlugin(Plugin):
    """Adds more powerful parsing of DATETIME fields.
    
    >>> parser.add_plugin(DateParserPlugin())
    >>> parser.parse(u"date:'last tuesday'")
    """
    
    def __init__(self, basedate=None, dateparser=None, callback=None,
                 free=False):
        """
        :param basedate: a datetime object representing the current time
            against which to measure relative dates. If you do not supply this
            argument, the plugin uses ``datetime.utcnow()``.
        :param dateparser: an instance of
            :class:`whoosh.qparser.dateparse.DateParser`. If you do not supply
            this argument, the plugin automatically uses
            :class:`whoosh.qparser.dateparse.English`.
        :param callback: a callback function for parsing errors. This allows
            you to provide feedback to the user about problems parsing dates.
        :param remove: if True, unparseable dates are removed from the token
            stream instead of being replaced with ErrorToken.
        :param free: if True, this plugin will install a filter early in the
            parsing process and try to find undelimited dates such as
            ``date:last tuesday``. Note that allowing this could result in
            normal query words accidentally being parsed as dates sometimes.
        """
        
        self.basedate = basedate
        if dateparser is None:
            dateparser = English()
        self.dateparser = dateparser
        self.callback = callback
        self.free = free
    
    def tokens(self, parser):
        if self.free:
            # If we're tokenizing, we have to go before the FieldsPlugin
            return ((DateToken, -1), )
        else:
            return ()
    
    def filters(self, parser):
        # Run the filter after the FieldsPlugin assigns field names
        return ((self.do_dates, 110), )
    
    def do_dates(self, parser, stream):
        schema = parser.schema
        if not schema:
            return stream
        
        from whoosh.fields import DATETIME
        datefields = frozenset(fieldname for fieldname, field
                               in parser.schema.items()
                               if isinstance(field, DATETIME))
        
        newstream = stream.empty()
        for t in stream:
            if isinstance(t, Group):
                t = self.do_dates(parser, t)
            elif (t.fieldname in datefields
                  or (t.fieldname is None and parser.fieldname in datefields)):
                if isinstance(t, Word):
                    text = t.text
                    try:
                        dt = self.dateparser.date_from(text, self.basedate)
                        if dt is None:
                            if self.callback:
                                self.callback(text)
                            t = ErrorToken(t)
                        else:
                            t = DateToken(t.fieldname, dt, t.boost)
                    except DateParseError, e:
                        if self.callback:
                            self.callback("%s (%r)" % (str(e), text))
                        t = ErrorToken(t)
                
                elif isinstance(t, RangePlugin.Range):
                    start = end = None
                    error = None
                    
                    dp = self.dateparser.get_parser()
                    
                    if t.start:
                        start = dp.date_from(t.start, self.basedate)
                        if start is None:
                            error = t.start
                    if t.end:
                        end = dp.date_from(t.end, self.basedate)
                        if end is None and error is None:
                            error = t.end
                    
                    if error is not None:
                        if self.callback:
                            self.callback(error)
                        t = ErrorToken(t)
                    else:
                        if start and end:
                            ts = timespan(start, end).disambiguated(self.basedate)
                            start, end = ts.start, ts.end
                        elif start:
                            start = start.disambiguated(self.basedate)
                            if isinstance(start, timespan):
                                start = start.start
                        elif end:
                            end = end.disambiguated(self.basedate)
                            if isinstance(end, timespan):
                                end = end.end
                        t = DateRangeToken(t.fieldname, start, end, boost=t.boost)
            
            newstream.append(t)
        return newstream


class DateToken(BasicSyntax):
    expr = re.compile("([A-Za-z][A-Za-z_0-9]*):([^^]+)")
    
    def __init__(self, fieldname, timeobj, boost=1.0, endpos=None):
        self.fieldname = fieldname
        self.timeobj = timeobj
        self.boost = boost
        self.endpos = endpos
    
    def __repr__(self):
        r = "%s:(%r)" % (self.fieldname, self.timeobj)
        if self.boost != 1.0:
            r + "^%s" % self.boost
        return r
    
    def set_boost(self, b):
        return self.__class__(self.fieldname, self.timeobj, boost=b,
                              endpos=self.endpos)
    
    def set_fieldname(self, name):
        if name is None:
            raise Exception
        return self.__class__(name, self.timeobj, boost=self.boost,
                              endpos=self.endpos)
    
    def query(self, parser):
        from whoosh import query
        
        fieldname = self.fieldname or parser.fieldname
        field = parser.schema[fieldname]
        dt = self.timeobj
        if isinstance(self.timeobj, datetime):
            return query.Term(fieldname, field.to_text(dt), boost=self.boost)
        elif isinstance(self.timeobj, timespan):
            return query.DateRange(fieldname, dt.start, dt.end,
                                   boost=self.boost)
        else:
            raise Exception("Unknown time object: %r" % dt)

    @classmethod
    def create(cls, parser, match):
        fieldname = match.group(1)
        if parser.schema and fieldname in parser.schema:
            field = parser.schema[fieldname]
            
            from whoosh.fields import DATETIME
            if isinstance(field, DATETIME):
                text = match.group(2)
                textstart = match.start(2)
                
                plugin = parser.get_plugin(DateParserPlugin)
                dateparser = plugin.dateparser
                basedate = plugin.basedate
                
                d, newpos = dateparser.parse(text, basedate)
                if d:
                    return cls(fieldname, d, endpos=newpos + textstart)
    

class DateRangeToken(BasicSyntax):
    def __init__(self, fieldname, starttime, endtime, boost=1.0, endpos=None):
        self.fieldname = fieldname
        self.starttime = starttime
        self.endtime = endtime
        self.boost = boost
        self.endpos = endpos
        
    def __repr__(self):
        r = "%s:(%r, %r)" % (self.fieldname, self.starttime, self.endtime)
        if self.boost != 1.0:
            r + "^%s" % self.boost
        return r
    
    def set_boost(self, b):
        return self.__class__(self.fieldname, self.starttime, self.endtime,
                              boost=b, endpos=self.endpos)
    
    def set_fieldname(self, name):
        if name is None:
            raise Exception
        return self.__class__(name, self.starttime, self.endtime,
                              boost=self.boost, endpos=self.endpos)
        
    def query(self, parser):
        from whoosh import query
        fieldname = self.fieldname or parser.fieldname
        start = self.starttime
        end = self.endtime
        
        if start is None and end is None:
            return query.Every(fieldname)
        else:
            return query.DateRange(fieldname, start, end, boost=self.boost)
    
    
    
    
    






from __future__ import with_statement

from nose.tools import assert_equal, assert_raises

from datetime import datetime, timedelta

from whoosh import analysis, fields, index, qparser, searching, scoring
from whoosh.filedb.filestore import RamStorage
from whoosh.query import *
from whoosh.util import permutations


def make_index():
    s = fields.Schema(key = fields.ID(stored = True),
                      name = fields.TEXT,
                      value = fields.TEXT)
    st = RamStorage()
    ix = st.create_index(s)
    
    w = ix.writer()
    w.add_document(key = u"A", name = u"Yellow brown", value = u"Blue red green render purple?")
    w.add_document(key = u"B", name = u"Alpha beta", value = u"Gamma delta epsilon omega.")
    w.add_document(key = u"C", name = u"One two", value = u"Three rendered four five.")
    w.add_document(key = u"D", name = u"Quick went", value = u"Every red town.")
    w.add_document(key = u"E", name = u"Yellow uptown", value = u"Interest rendering outer photo!")
    w.commit()
    
    return ix

def _get_keys(stored_fields):
    return sorted([d.get("key") for d in stored_fields])

def _docs(q, s):
    return _get_keys([s.stored_fields(docnum) for docnum
                           in q.docs(s)])

def _run_query(q, target):
    ix = make_index()
    with ix.searcher() as s:
        assert_equal(target, _docs(q, s))

def test_empty_index():
    schema = fields.Schema(key = fields.ID(stored=True), value = fields.TEXT)
    st = RamStorage()
    assert_raises(index.EmptyIndexError, st.open_index, schema=schema)

def test_docs_method():
    ix = make_index()
    with ix.searcher() as s:
        assert_equal(_get_keys(s.documents(name="yellow")), [u"A", u"E"])
        assert_equal(_get_keys(s.documents(value="red")), [u"A", u"D"])

def test_term():
    _run_query(Term("name", u"yellow"), [u"A", u"E"])
    _run_query(Term("value", u"zeta"), [])
    _run_query(Term("value", u"red"), [u"A", u"D"])
    
def test_require():
    _run_query(Require(Term("value", u"red"), Term("name", u"yellow")),
                    [u"A"])
    
def test_and():
    _run_query(And([Term("value", u"red"), Term("name", u"yellow")]),
                    [u"A"])
    # Missing
    _run_query(And([Term("value", u"ochre"), Term("name", u"glonk")]),
                    [])
    
def test_or():
    _run_query(Or([Term("value", u"red"), Term("name", u"yellow")]),
                    [u"A", u"D", u"E"])
    # Missing
    _run_query(Or([Term("value", u"ochre"), Term("name", u"glonk")]),
                    [])
    _run_query(Or([]), [])

def test_not():
    _run_query(Or([Term("value", u"red"), Term("name", u"yellow"), Not(Term("name", u"quick"))]),
                    [u"A", u"E"])

def test_topnot():
    _run_query(Not(Term("value", "red")), [u"B", "C", "E"])
    _run_query(Not(Term("name", "yellow")), [u"B", u"C", u"D"])

def test_andnot():
    _run_query(AndNot(Term("name", u"yellow"), Term("value", u"purple")),
                    [u"E"])

def test_variations():
    _run_query(Variations("value", u"render"), [u"A", u"C", u"E"])

def test_wildcard():
    _run_query(Or([Wildcard('value', u'*red*'), Wildcard('name', u'*yellow*')]),
                    [u"A", u"C", u"D", u"E"])
    # Missing
    _run_query(Wildcard('value', 'glonk*'), [])

def test_not2():
    schema = fields.Schema(name=fields.ID(stored=True), value=fields.TEXT)
    storage = RamStorage()
    ix = storage.create_index(schema)
    writer = ix.writer()
    writer.add_document(name=u"a", value=u"alfa bravo charlie delta echo")
    writer.add_document(name=u"b", value=u"bravo charlie delta echo foxtrot")
    writer.add_document(name=u"c", value=u"charlie delta echo foxtrot golf")
    writer.add_document(name=u"d", value=u"delta echo golf hotel india")
    writer.add_document(name=u"e", value=u"echo golf hotel india juliet")
    writer.commit()
    
    with ix.searcher() as s:
        p = qparser.QueryParser("value", None)
        results = s.search(p.parse("echo NOT golf"))
        assert_equal(sorted([d["name"] for d in results]), ["a", "b"])
        
        results = s.search(p.parse("echo NOT bravo"))
        assert_equal(sorted([d["name"] for d in results]), ["c", "d", "e"])
    
    ix.delete_by_term("value", u"bravo")
    
    with ix.searcher() as s:
        results = s.search(p.parse("echo NOT charlie"))
        assert_equal(sorted([d["name"] for d in results]), ["d", "e"])

#    def test_or_minmatch():
#        schema = fields.Schema(k=fields.STORED, v=fields.TEXT)
#        st = RamStorage()
#        ix = st.create_index(schema)
#        
#        w = ix.writer()
#        w.add_document(k=1, v=u"alfa bravo charlie delta echo")
#        w.add_document(k=2, v=u"bravo charlie delta echo foxtrot")
#        w.add_document(k=3, v=u"charlie delta echo foxtrot golf")
#        w.add_document(k=4, v=u"delta echo foxtrot golf hotel")
#        w.add_document(k=5, v=u"echo foxtrot golf hotel india")
#        w.add_document(k=6, v=u"foxtrot golf hotel india juliet")
#        w.commit()
#        
#        s = ix.searcher()
#        q = Or([Term("v", "echo"), Term("v", "foxtrot")], minmatch=2)
#        r = s.search(q)
#        assert sorted(d["k"] for d in r), [2, 3, 4, 5])

def test_range():
    schema = fields.Schema(id=fields.ID(stored=True), content=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"A", content=u"alfa bravo charlie delta echo")
    w.add_document(id=u"B", content=u"bravo charlie delta echo foxtrot")
    w.add_document(id=u"C", content=u"charlie delta echo foxtrot golf")
    w.add_document(id=u"D", content=u"delta echo foxtrot golf hotel")
    w.add_document(id=u"E", content=u"echo foxtrot golf hotel india")
    w.commit()
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("content", schema)
        
        q = qp.parse(u"charlie [delta TO foxtrot]")
        assert_equal(q.__class__, And)
        assert_equal(q[0].__class__, Term)
        assert_equal(q[1].__class__, TermRange)
        assert_equal(q[1].start, "delta")
        assert_equal(q[1].end, "foxtrot")
        assert_equal(q[1].startexcl, False)
        assert_equal(q[1].endexcl, False)
        ids = sorted([d['id'] for d in s.search(q)])
        assert_equal(ids, [u'A', u'B', u'C'])
        
        q = qp.parse(u"foxtrot {echo TO hotel]")
        assert_equal(q.__class__, And)
        assert_equal(q[0].__class__, Term)
        assert_equal(q[1].__class__, TermRange)
        assert_equal(q[1].start, "echo")
        assert_equal(q[1].end, "hotel")
        assert_equal(q[1].startexcl, True)
        assert_equal(q[1].endexcl, False)
        ids = sorted([d['id'] for d in s.search(q)])
        assert_equal(ids, [u'B', u'C', u'D', u'E'])
        
        q = qp.parse(u"{bravo TO delta}")
        assert_equal(q.__class__, TermRange)
        assert_equal(q.start, "bravo")
        assert_equal(q.end, "delta")
        assert_equal(q.startexcl, True)
        assert_equal(q.endexcl, True)
        ids = sorted([d['id'] for d in s.search(q)])
        assert_equal(ids, [u'A', u'B', u'C'])
        
        # Shouldn't match anything
        q = qp.parse(u"[1 to 10]")
        assert_equal(q.__class__, TermRange)
        assert_equal(len(s.search(q)), 0)

def test_range_clusiveness():
    schema = fields.Schema(id=fields.ID(stored=True))
    st = RamStorage()
    ix = st.create_index(schema)
    w = ix.writer()
    for letter in u"abcdefg":
        w.add_document(id=letter)
    w.commit()
    
    with ix.searcher() as s:
        def check(startexcl, endexcl, string):
            q = TermRange("id", "b", "f", startexcl, endexcl)
            r = "".join(sorted(d['id'] for d in s.search(q)))
            assert_equal(r, string)
            
        check(False, False, "bcdef")
        check(True, False, "cdef")
        check(True, True, "cde")
        check(False, True, "bcde")
    
def test_open_ranges():
    schema = fields.Schema(id=fields.ID(stored=True))
    st = RamStorage()
    ix = st.create_index(schema)
    w = ix.writer()
    for letter in u"abcdefg":
        w.add_document(id=letter)
    w.commit()
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("id", schema)
        def check(qstring, result):
            q = qp.parse(qstring)
            r = "".join(sorted([d['id'] for d in s.search(q)]))
            assert_equal(r, result)
            
        check(u"[b TO]", "bcdefg")
        check(u"[TO e]", "abcde")
        check(u"[b TO d]", "bcd")
        check(u"{b TO]", "cdefg")
        check(u"[TO e}", "abcd")
        check(u"{b TO d}", "c")

def test_open_numeric_ranges():
    domain = range(0, 10000, 7)
    
    schema = fields.Schema(num=fields.NUMERIC(stored=True))
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    for i in domain:
        w.add_document(num=i)
    w.commit()
    
    qp = qparser.QueryParser("num", schema)
    with ix.searcher() as s:
        q = qp.parse("[100 to]")
        r = [hit["num"] for hit in s.search(q, limit=None)]
        assert_equal(r, [n for n in domain if n >= 100])
        
        q = qp.parse("[to 5000]")
        r = [hit["num"] for hit in s.search(q, limit=None)]
        assert_equal(r, [n for n in domain if n <= 5000])

def test_open_date_ranges():
    basedate = datetime(2011, 1, 24, 6, 25, 0, 0)
    domain = [basedate + timedelta(days=n) for n in xrange(-20, 20)]
    
    schema = fields.Schema(date=fields.DATETIME(stored=True))
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    for d in domain:
        w.add_document(date=d)
    w.commit()
    
    with ix.searcher() as s:
        # Without date parser
        qp = qparser.QueryParser("date", schema)
        q = qp.parse("[2011-01-10 to]")
        r = [hit["date"] for hit in s.search(q, limit=None)]
        assert len(r) > 0
        target = [d for d in domain if d >= datetime(2011, 1, 10, 6, 25)]
        assert_equal(r, target)
        
        q = qp.parse("[to 2011-01-30]")
        r = [hit["date"] for hit in s.search(q, limit=None)]
        assert len(r) > 0
        target = [d for d in domain if d <= datetime(2011, 1, 30, 6, 25)]
        assert_equal(r, target)
    
        # With date parser
        from whoosh.qparser.dateparse import DateParserPlugin
        qp.add_plugin(DateParserPlugin(basedate))
        
        q = qp.parse("[10 jan 2011 to]")
        r = [hit["date"] for hit in s.search(q, limit=None)]
        assert len(r) > 0
        target = [d for d in domain if d >= datetime(2011, 1, 10, 6, 25)]
        assert_equal(r, target)
        
        q = qp.parse("[to 30 jan 2011]")
        r = [hit["date"] for hit in s.search(q, limit=None)]
        assert len(r) > 0
        target = [d for d in domain if d <= datetime(2011, 1, 30, 6, 25)]
        assert_equal(r, target)
        
def test_negated_unlimited_ranges():
    # Whoosh should treat u"[to]" as if it was "*"
    schema = fields.Schema(id=fields.ID(stored=True), num=fields.NUMERIC,
                           date=fields.DATETIME)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    from string import ascii_letters
    domain = unicode(ascii_letters)
    
    dt = datetime.now()
    for i, letter in enumerate(domain):
        w.add_document(id=letter, num=i, date=dt + timedelta(days=i))
    w.commit()
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("id", schema)
        
        nq = qp.parse(u"NOT [to]")
        assert_equal(nq.__class__, Not)
        q = nq.query
        assert_equal(q.__class__, Every)
        assert_equal("".join(h["id"] for h in s.search(q, limit=None)), domain)
        assert_equal(list(nq.docs(s)), [])
        
        nq = qp.parse(u"NOT num:[to]")
        assert_equal(nq.__class__, Not)
        q = nq.query
        assert_equal(q.__class__, NumericRange)
        assert_equal(q.start, None)
        assert_equal(q.end, None)
        assert_equal("".join(h["id"] for h in s.search(q, limit=None)), domain)
        assert_equal(list(nq.docs(s)), [])
        
        nq = qp.parse(u"NOT date:[to]")
        assert_equal(nq.__class__, Not)
        q = nq.query
        assert_equal(q.__class__, Every)
        assert_equal("".join(h["id"] for h in s.search(q, limit=None)), domain)
        assert_equal(list(nq.docs(s)), [])

def test_keyword_or():
    schema = fields.Schema(a=fields.ID(stored=True), b=fields.KEYWORD)
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(a=u"First", b=u"ccc ddd")
    w.add_document(a=u"Second", b=u"aaa ddd")
    w.add_document(a=u"Third", b=u"ccc eee")
    w.commit()
    
    qp = qparser.QueryParser("b", schema)
    with ix.searcher() as s:
        qr = qp.parse(u"b:ccc OR b:eee")
        assert_equal(qr.__class__, Or)
        r = s.search(qr)
        assert_equal(len(r), 2)
        assert_equal(r[0]["a"], "Third")
        assert_equal(r[1]["a"], "First")

def test_merged():
    sc = fields.Schema(id=fields.ID(stored=True), content=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(sc)
    w = ix.writer()
    w.add_document(id=u"alfa", content=u"alfa")
    w.add_document(id=u"bravo", content=u"bravo")
    w.add_document(id=u"charlie", content=u"charlie")
    w.add_document(id=u"delta", content=u"delta")
    w.commit()
    
    with ix.searcher() as s:
        r = s.search(Term("content", u"bravo"))
        assert_equal(len(r), 1)
        assert_equal(r[0]["id"], "bravo")
    
    w = ix.writer()
    w.add_document(id=u"echo", content=u"echo")
    w.commit()
    assert_equal(len(ix._segments()), 1)
    
    with ix.searcher() as s:
        r = s.search(Term("content", u"bravo"))
        assert_equal(len(r), 1)
        assert_equal(r[0]["id"], "bravo")
    
def test_multireader():
    sc = fields.Schema(id=fields.ID(stored=True), content=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(sc)
    w = ix.writer()
    w.add_document(id=u"alfa", content=u"alfa")
    w.add_document(id=u"bravo", content=u"bravo")
    w.add_document(id=u"charlie", content=u"charlie")
    w.add_document(id=u"delta", content=u"delta")
    w.add_document(id=u"echo", content=u"echo")
    w.add_document(id=u"foxtrot", content=u"foxtrot")
    w.add_document(id=u"golf", content=u"golf")
    w.add_document(id=u"hotel", content=u"hotel")
    w.add_document(id=u"india", content=u"india")
    w.commit()
    
    with ix.searcher() as s:
        r = s.search(Term("content", u"bravo"))
        assert_equal(len(r), 1)
        assert_equal(r[0]["id"], "bravo")
    
    w = ix.writer()
    w.add_document(id=u"juliet", content=u"juliet")
    w.add_document(id=u"kilo", content=u"kilo")
    w.add_document(id=u"lima", content=u"lima")
    w.add_document(id=u"mike", content=u"mike")
    w.add_document(id=u"november", content=u"november")
    w.add_document(id=u"oscar", content=u"oscar")
    w.add_document(id=u"papa", content=u"papa")
    w.add_document(id=u"quebec", content=u"quebec")
    w.add_document(id=u"romeo", content=u"romeo")
    w.commit()
    assert_equal(len(ix._segments()), 2)
    
    #r = ix.reader()
    #assert r.__class__.__name__, "MultiReader")
    #pr = r.postings("content", u"bravo")
    
    with ix.searcher() as s:
        r = s.search(Term("content", u"bravo"))
        assert_equal(len(r), 1)
        assert_equal(r[0]["id"], "bravo")

def test_posting_phrase():
    schema = fields.Schema(name=fields.ID(stored=True), value=fields.TEXT)
    storage = RamStorage()
    ix = storage.create_index(schema)
    writer = ix.writer()
    writer.add_document(name=u"A", value=u"Little Miss Muffet sat on a tuffet")
    writer.add_document(name=u"B", value=u"Miss Little Muffet tuffet")
    writer.add_document(name=u"C", value=u"Miss Little Muffet tuffet sat")
    writer.add_document(name=u"D", value=u"Gibberish blonk falunk miss muffet sat tuffet garbonzo")
    writer.add_document(name=u"E", value=u"Blah blah blah pancakes")
    writer.commit()
    
    with ix.searcher() as s:
        def names(results):
            return sorted([fields['name'] for fields in results])
        
        q = Phrase("value", [u"little", u"miss", u"muffet", u"sat", u"tuffet"])
        m = q.matcher(s)
        assert_equal(m.__class__.__name__, "SpanNearMatcher")
        
        r = s.search(q)
        assert_equal(names(r), ["A"])
        assert_equal(len(r), 1)
        
        q = Phrase("value", [u"miss", u"muffet", u"sat", u"tuffet"])
        assert_equal(names(s.search(q)), ["A", "D"])
        
        q = Phrase("value", [u"falunk", u"gibberish"])
        r = s.search(q)
        assert_equal(names(r), [])
        assert_equal(len(r), 0)
        
        q = Phrase("value", [u"gibberish", u"falunk"], slop=2)
        assert_equal(names(s.search(q)), ["D"])
        
        q = Phrase("value", [u"blah"] * 4)
        assert_equal(names(s.search(q)), [])  # blah blah blah blah
        
        q = Phrase("value", [u"blah"] * 3)
        m = q.matcher(s)
        assert_equal(names(s.search(q)), ["E"])

#    def test_vector_phrase():
#        ana = analysis.StandardAnalyzer()
#        ftype = fields.FieldType(format=formats.Frequency(ana),
#                                 vector=formats.Positions(ana),
#                                 scorable=True)
#        schema = fields.Schema(name=fields.ID(stored=True), value=ftype)
#        storage = RamStorage()
#        ix = storage.create_index(schema)
#        writer = ix.writer()
#        writer.add_document(name=u"A", value=u"Little Miss Muffet sat on a tuffet")
#        writer.add_document(name=u"B", value=u"Miss Little Muffet tuffet")
#        writer.add_document(name=u"C", value=u"Miss Little Muffet tuffet sat")
#        writer.add_document(name=u"D", value=u"Gibberish blonk falunk miss muffet sat tuffet garbonzo")
#        writer.add_document(name=u"E", value=u"Blah blah blah pancakes")
#        writer.commit()
#        
#        searcher = ix.searcher()
#        
#        def names(results):
#            return sorted([fields['name'] for fields in results])
#        
#        q = Phrase("value", [u"little", u"miss", u"muffet", u"sat", u"tuffet"])
#        m = q.matcher(searcher)
#        assert m.__class__.__name__, "VectorPhraseMatcher")
#        
#        assert names(searcher.search(q)), ["A"])
#        
#        q = Phrase("value", [u"miss", u"muffet", u"sat", u"tuffet"])
#        assert names(searcher.search(q)), ["A", "D"])
#        
#        q = Phrase("value", [u"falunk", u"gibberish"])
#        assert names(searcher.search(q)), [])
#        
#        q = Phrase("value", [u"gibberish", u"falunk"], slop=2)
#        assert names(searcher.search(q)), ["D"])
#        
#        #q = Phrase("value", [u"blah"] * 4)
#        #assert names(searcher.search(q)), []) # blah blah blah blah
#        
#        q = Phrase("value", [u"blah"] * 3)
#        assert names(searcher.search(q)), ["E"])
    
def test_phrase_score():
    schema = fields.Schema(name=fields.ID(stored=True), value=fields.TEXT)
    storage = RamStorage()
    ix = storage.create_index(schema)
    writer = ix.writer()
    writer.add_document(name=u"A", value=u"Little Miss Muffet sat on a tuffet")
    writer.add_document(name=u"D", value=u"Gibberish blonk falunk miss muffet sat tuffet garbonzo")
    writer.add_document(name=u"E", value=u"Blah blah blah pancakes")
    writer.add_document(name=u"F", value=u"Little miss muffet little miss muffet")
    writer.commit()
    
    with ix.searcher() as s:
        q = Phrase("value", [u"little", u"miss", u"muffet"])
        m = q.matcher(s)
        assert_equal(m.id(), 0)
        score1 = m.weight()
        assert score1 > 0
        m.next()
        assert_equal(m.id(), 3)
        assert m.weight() > score1

def test_stop_phrase():
    schema = fields.Schema(title=fields.TEXT(stored=True))
    storage = RamStorage()
    ix = storage.create_index(schema)
    writer = ix.writer()
    writer.add_document(title=u"Richard of York")
    writer.add_document(title=u"Lily the Pink")
    writer.commit()
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("title", schema)
        q = qp.parse(u"richard of york")
        assert_equal(len(s.search(q)), 1)
        #q = qp.parse(u"lily the pink")
        #assert len(s.search(q)), 1)
        assert_equal(len(s.find("title", u"lily the pink")), 1)

def test_phrase_order():
    tfield = fields.TEXT(stored=True, analyzer=analysis.SimpleAnalyzer())
    schema = fields.Schema(text=tfield)
    storage = RamStorage()
    ix = storage.create_index(schema)
    
    writer = ix.writer()
    for ls in permutations(["ape", "bay", "can", "day"], 4):
        writer.add_document(text=u" ".join(ls))
    writer.commit()
    
    with ix.searcher() as s:
        def result(q):
            r = s.search(q, limit=None, sortedby=None)
            return sorted([d['text'] for d in r])
        
        q = Phrase("text", ["bay", "can", "day"])
        assert_equal(result(q), [u'ape bay can day', u'bay can day ape'])
    
def test_phrase_sameword():
    schema = fields.Schema(id=fields.STORED, text=fields.TEXT)
    storage = RamStorage()
    ix = storage.create_index(schema)
    
    writer = ix.writer()
    writer.add_document(id=1, text=u"The film Linda Linda Linda is good")
    writer.add_document(id=2, text=u"The model Linda Evangelista is pretty")
    writer.commit()
    
    with ix.searcher() as s:
        r = s.search(Phrase("text", ["linda", "linda", "linda"]), limit=None)
        assert_equal(len(r), 1)
        assert_equal(r[0]["id"], 1)

def test_phrase_multi():
    schema = fields.Schema(id=fields.STORED, text=fields.TEXT)
    ix = RamStorage().create_index(schema)
    
    domain = u"alfa bravo charlie delta echo".split()
    w = None
    for i, ls in enumerate(permutations(domain)):
        if w is None:
            w = ix.writer()
        w.add_document(id=i, text=u" ".join(ls))
        if not i % 30:
            w.commit()
            w = None
    if w is not None:
        w.commit()
    
    with ix.searcher() as s:
        q = Phrase("text", ["alfa", "bravo"])
        r = s.search(q)

def test_missing_field_scoring():
    schema = fields.Schema(name=fields.TEXT(stored=True),
                           hobbies=fields.TEXT(stored=True))
    storage = RamStorage()
    ix = storage.create_index(schema)
    writer = ix.writer() 
    writer.add_document(name=u'Frank', hobbies=u'baseball, basketball')
    writer.commit()
    r = ix.reader()
    assert_equal(r.field_length("hobbies"), 2)
    assert_equal(r.field_length("name"), 1)
    r.close()
    
    writer = ix.writer()
    writer.add_document(name=u'Jonny') 
    writer.commit()
    
    with ix.searcher() as s:
        r = s.reader()
        assert_equal(len(ix._segments()), 1)
        assert_equal(r.field_length("hobbies"), 2)
        assert_equal(r.field_length("name"), 2)
        
        parser = qparser.MultifieldParser(['name', 'hobbies'], schema)
        q = parser.parse(u"baseball")
        result = s.search(q)
        assert_equal(len(result), 1)
    
def test_search_fieldname_underscores():
    s = fields.Schema(my_name=fields.ID(stored=True), my_value=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(s)
    
    w = ix.writer()
    w.add_document(my_name=u"Green", my_value=u"It's not easy being green")
    w.add_document(my_name=u"Red", my_value=u"Hopping mad like a playground ball")
    w.commit()
    
    qp = qparser.QueryParser("my_value", schema=s)
    with ix.searcher() as s:
        r = s.search(qp.parse(u"my_name:Green"))
        assert_equal(r[0]['my_name'], "Green")
    
def test_short_prefix():
    s = fields.Schema(name=fields.ID, value=fields.TEXT)
    qp = qparser.QueryParser("value", schema=s)
    q = qp.parse(u"s*")
    assert_equal(q.__class__.__name__, "Prefix")
    assert_equal(q.text, "s")
    
def test_weighting():
    from whoosh.scoring import Weighting, BaseScorer
    
    schema = fields.Schema(id=fields.ID(stored=True),
                           n_comments=fields.STORED)
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1", n_comments=5)
    w.add_document(id=u"2", n_comments=12)
    w.add_document(id=u"3", n_comments=2)
    w.add_document(id=u"4", n_comments=7)
    w.commit()
    
    # Fake Weighting implementation
    class CommentWeighting(Weighting):
        def scorer(self, searcher, fieldname, text, qf=1):
            return self.CommentScorer(searcher.stored_fields)
        
        class CommentScorer(BaseScorer):
            def __init__(self, stored_fields):
                self.stored_fields = stored_fields
        
            def score(self, matcher):
                ncomments = self.stored_fields(matcher.id()).get("n_comments", 0)
                return ncomments
    
    with ix.searcher(weighting=CommentWeighting()) as s:
        q = TermRange("id", u"1", u"4", constantscore=False)
        
        r = s.search(q)
        ids = [fs["id"] for fs in r]
        assert_equal(ids, ["2", "4", "1", "3"])

def test_dismax():
    schema = fields.Schema(id=fields.STORED,
                           f1=fields.TEXT, f2=fields.TEXT, f3=fields.TEXT)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=1, f1=u"alfa bravo charlie delta",
                   f2=u"alfa alfa alfa",
                   f3 = u"alfa echo foxtrot hotel india")
    w.commit()
    
    with ix.searcher(weighting=scoring.Frequency()) as s:
        assert_equal(list(s.documents(f1="alfa")), [{"id": 1}])
        assert_equal(list(s.documents(f2="alfa")), [{"id": 1}])
        assert_equal(list(s.documents(f3="alfa")), [{"id": 1}])
        
        qs = [Term("f1", "alfa"), Term("f2", "alfa"), Term("f3", "alfa")]
        dm = DisjunctionMax(qs)
        r = s.search(dm)
        assert_equal(r.score(0), 3.0)

def test_deleted_wildcard():
    schema = fields.Schema(id=fields.ID(stored=True))
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"alfa")
    w.add_document(id=u"bravo")
    w.add_document(id=u"charlie")
    w.add_document(id=u"delta")
    w.add_document(id=u"echo")
    w.add_document(id=u"foxtrot")
    w.commit()
    
    w = ix.writer()
    w.delete_by_term("id", "bravo")
    w.delete_by_term("id", "delta")
    w.delete_by_term("id", "echo")
    w.commit()
    
    with ix.searcher() as s:
        r = s.search(Every("id"))
        assert_equal(sorted([d['id'] for d in r]), ["alfa", "charlie", "foxtrot"])
    
def test_missing_wildcard():
    schema = fields.Schema(id=fields.ID(stored=True), f1=fields.TEXT, f2=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1", f1=u"alfa", f2=u"apple")
    w.add_document(id=u"2", f1=u"bravo")
    w.add_document(id=u"3", f1=u"charlie", f2=u"candy")
    w.add_document(id=u"4", f2=u"donut")
    w.add_document(id=u"5")
    w.commit()
    
    with ix.searcher() as s:
        r = s.search(Every("id"))
        assert_equal(sorted([d['id'] for d in r]), ["1", "2", "3", "4", "5"])
        
        r = s.search(Every("f1"))
        assert_equal(sorted([d['id'] for d in r]), ["1", "2", "3"])
        
        r = s.search(Every("f2"))
        assert_equal(sorted([d['id'] for d in r]), ["1", "3", "4"])

def test_finalweighting():
    from whoosh.scoring import Frequency
    
    schema = fields.Schema(id=fields.ID(stored=True),
                           summary=fields.TEXT,
                           n_comments=fields.STORED)
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1", summary=u"alfa bravo", n_comments=5)
    w.add_document(id=u"2", summary=u"alfa", n_comments=12)
    w.add_document(id=u"3", summary=u"bravo", n_comments=2)
    w.add_document(id=u"4", summary=u"bravo bravo", n_comments=7)
    w.commit()
    
    class CommentWeighting(Frequency):
        use_final = True
        
        def final(self, searcher, docnum, score):
            ncomments = searcher.stored_fields(docnum).get("n_comments", 0)
            return ncomments
    
    with ix.searcher(weighting=CommentWeighting()) as s:
        r = s.search(qparser.QueryParser("summary", None).parse("alfa OR bravo"))
        ids = [fs["id"] for fs in r]
        assert_equal(["2", "4", "1", "3"], ids)
    
def test_outofdate():
    schema = fields.Schema(id=fields.ID(stored=True))
    st = RamStorage()
    ix = st.create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1")
    w.add_document(id=u"2")
    w.commit()
    
    s = ix.searcher()
    assert s.up_to_date()
    
    w = ix.writer()
    w.add_document(id=u"3")
    w.add_document(id=u"4")
    
    assert s.up_to_date()
    w.commit()
    assert not s.up_to_date()

    s = s.refresh()
    assert s.up_to_date()
    s.close()

def test_find_missing():
    schema = fields.Schema(id=fields.ID, text=fields.KEYWORD(stored=True))
    ix = RamStorage().create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1", text=u"alfa")
    w.add_document(id=u"2", text=u"bravo")
    w.add_document(text=u"charlie")
    w.add_document(id=u"4", text=u"delta")
    w.add_document(text=u"echo")
    w.add_document(id=u"6", text=u"foxtrot")
    w.add_document(text=u"golf")
    w.commit()
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("text", schema)
        q = qp.parse(u"NOT id:*")
        r = s.search(q, limit=None)
        assert_equal(list(h["text"] for h in r), ["charlie", "echo", "golf"])

def test_ngram_phrase():
    schema = fields.Schema(text=fields.NGRAM(minsize=2, maxsize=2, phrase=True), path=fields.ID(stored=True)) 
    ix = RamStorage().create_index(schema)
    writer = ix.writer()
    writer.add_document(text=u'\u9AD8\u6821\u307E\u3067\u306F\u6771\u4EAC\u3067\u3001\u5927\u5B66\u304B\u3089\u306F\u4EAC\u5927\u3067\u3059\u3002', path=u'sample') 
    writer.commit()
    
    with ix.searcher() as s:
        p = qparser.QueryParser("text", schema)
        
        q = p.parse(u'\u6771\u4EAC\u5927\u5B66')
        assert_equal(len(s.search(q)), 1)
        
        q = p.parse(u'"\u6771\u4EAC\u5927\u5B66"')
        assert_equal(len(s.search(q)), 0)
        
        q = p.parse(u'"\u306F\u6771\u4EAC\u3067"')
        assert_equal(len(s.search(q)), 1)
    
def test_ordered():
    domain = u"alfa bravo charlie delta echo foxtrot".split(" ")
    
    schema = fields.Schema(f=fields.TEXT(stored=True))
    ix = RamStorage().create_index(schema)
    writer = ix.writer()
    for ls in permutations(domain):
        writer.add_document(f=u" ".join(ls))
    writer.commit()
    
    with ix.searcher() as s:
        q = Ordered([Term("f", u"alfa"), Term("f", u"charlie"), Term("f", "echo")])
        r = s.search(q)
        for hit in r:
            ls = hit["f"].split()
            assert "alfa" in ls
            assert "charlie" in ls
            assert "echo" in ls
            a = ls.index("alfa")
            c = ls.index("charlie")
            e = ls.index("echo")
            assert a < c and c < e, repr(ls)

def test_otherwise():
    schema = fields.Schema(id=fields.STORED, f=fields.TEXT)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=1, f=u"alfa one two")
    w.add_document(id=2, f=u"alfa three four")
    w.add_document(id=3, f=u"bravo four five")
    w.add_document(id=4, f=u"bravo six seven")
    w.commit()
    
    with ix.searcher() as s:
        q = Otherwise(Term("f", u"alfa"), Term("f", u"six"))
        assert_equal([d["id"] for d in s.search(q)], [1, 2])
        
        q = Otherwise(Term("f", u"tango"), Term("f", u"four"))
        assert_equal([d["id"] for d in s.search(q)], [2, 3])
        
        q = Otherwise(Term("f", u"tango"), Term("f", u"nine"))
        assert_equal([d["id"] for d in s.search(q)], [])

def test_fuzzyterm():
    schema = fields.Schema(id=fields.STORED, f=fields.TEXT)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=1, f=u"alfa bravo charlie delta")
    w.add_document(id=2, f=u"bravo charlie delta echo")
    w.add_document(id=3, f=u"charlie delta echo foxtrot")
    w.add_document(id=4, f=u"delta echo foxtrot golf")
    w.commit()
    
    with ix.searcher() as s:
        q = FuzzyTerm("f", "brave")
        assert_equal([d["id"] for d in s.search(q)], [1, 2])
    
def test_multireader_not():
    schema = fields.Schema(id=fields.STORED, f=fields.TEXT)
    
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=0, f=u"alfa bravo chralie")
    w.add_document(id=1, f=u"bravo chralie delta")
    w.add_document(id=2, f=u"charlie delta echo")
    w.add_document(id=3, f=u"delta echo foxtrot")
    w.add_document(id=4, f=u"echo foxtrot golf")
    w.commit()
    
    with ix.searcher() as s:
        q = And([Term("f", "delta"), Not(Term("f", "delta"))])
        r = s.search(q)
        assert_equal(len(r), 0)
    
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=5, f=u"alfa bravo chralie")
    w.add_document(id=6, f=u"bravo chralie delta")
    w.commit(merge=False)
    w = ix.writer()
    w.add_document(id=7, f=u"charlie delta echo")
    w.add_document(id=8, f=u"delta echo foxtrot")
    w.commit(merge=False)
    w = ix.writer()
    w.add_document(id=9, f=u"echo foxtrot golf")
    w.add_document(id=10, f=u"foxtrot golf delta")
    w.commit(merge=False)
    assert len(ix._segments()) > 1
    
    with ix.searcher() as s:
        q = And([Term("f", "delta"), Not(Term("f", "delta"))])
        r = s.search(q)
        assert_equal(len(r), 0)

def test_boost_phrase():
    schema = fields.Schema(title=fields.TEXT(field_boost=5.0, stored=True), text=fields.TEXT)
    ix = RamStorage().create_index(schema)
    domain = u"alfa bravo charlie delta".split()
    w = ix.writer()
    for ls in permutations(domain):
        t = u" ".join(ls)
        w.add_document(title=t, text=t)
    w.commit()
    
    q = Or([Term("title", u"alfa"), Term("title", u"bravo"), Phrase("text", [u"bravo", u"charlie", u"delta"])])
    
    def boost_phrases(q):
        if isinstance(q, Phrase):
            q.boost *= 1000.0
            return q
        else:
            return q.apply(boost_phrases)
    q = boost_phrases(q)
    
    with ix.searcher() as s:
        r = s.search(q, limit=None)
        for hit in r:
            if "bravo charlie delta" in hit["title"]:
                assert hit.score > 100.0

def test_trackingcollector():
    schema = fields.Schema(text=fields.TEXT(stored=True))
    ix = RamStorage().create_index(schema)
    domain = u"alfa bravo charlie delta echo".split()
    w = ix.writer()
    for ls in list(permutations(domain, 3))[::2]:
        w.add_document(text=u" ".join(ls))
    w.commit()
    
    with ix.searcher() as s:
        q = Or([Term("text", u"alfa"),Term("text", u"bravo"),
                Not(Term("text", "charlie"))])
        
        col = searching.TermTrackingCollector()
        r = s.search(q, collector=col)
        
        for docnum in col.catalog["text:alfa"]:
            words = s.stored_fields(docnum)["text"].split()
            assert "alfa" in words
            assert "charlie" not in words
        
        for docnum in col.catalog["text:bravo"]:
            words = s.stored_fields(docnum)["text"].split()
            assert "bravo" in words
            assert "charlie" not in words

def test_filter():
    schema = fields.Schema(id=fields.STORED, path=fields.ID, text=fields.TEXT)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=1, path=u"/a/1", text=u"alfa bravo charlie")
    w.add_document(id=2, path=u"/b/1", text=u"bravo charlie delta")
    w.add_document(id=3, path=u"/c/1", text=u"charlie delta echo")
    w.commit(merge=False)
    w = ix.writer()
    w.add_document(id=4, path=u"/a/2", text=u"delta echo alfa")
    w.add_document(id=5, path=u"/b/2", text=u"echo alfa bravo")
    w.add_document(id=6, path=u"/c/2", text=u"alfa bravo charlie")
    w.commit(merge=False)
    w = ix.writer()
    w.add_document(id=7, path=u"/a/3", text=u"bravo charlie delta")
    w.add_document(id=8, path=u"/b/3", text=u"charlie delta echo")
    w.add_document(id=9, path=u"/c/3", text=u"delta echo alfa")
    w.commit(merge=False)
    
    with ix.searcher() as s:
        fq = Or([Prefix("path", "/a"), Prefix("path", "/b")])
        r = s.search(Term("text", "alfa"), scored=False, filter=fq)
        assert_equal([d["id"] for d in r], [1, 4, 5])
        
        r = s.search(Term("text", "bravo"), scored=False, filter=fq)
        assert_equal([d["id"] for d in r], [1, 2, 5, 7,])
    
def test_timelimit():
    schema = fields.Schema(text=fields.TEXT)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    for _ in xrange(50):
        w.add_document(text=u"alfa")
    w.commit()
    
    import time
    from whoosh import matching
    
    class SlowMatcher(matching.WrappingMatcher):
        def next(self):
            time.sleep(0.02)
            self.child.next()
    
    class SlowQuery(WrappingQuery):
        def matcher(self, searcher):
            return SlowMatcher(self.child.matcher(searcher))
    
    with ix.searcher() as s:
        oq = Term("text", u"alfa")
        sq = SlowQuery(oq)
        
        col = searching.Collector(timelimit=0.1)
        assert_raises(searching.TimeLimit, s.search, sq, limit=None, collector=col)
        
        col = searching.Collector(timelimit=0.1)
        assert_raises(searching.TimeLimit, s.search, sq, limit=40, collector=col)
        
        col = searching.Collector(timelimit=0.25)
        try:
            s.search(sq, limit=None, collector=col)
        except searching.TimeLimit:
            r = col.results()
            assert r.scored_length() > 0
        
        col = searching.Collector(timelimit=0.5)
        r = s.search(oq, limit=None, collector=col)
        assert r.runtime < 0.5
            
def test_fieldboost():
    schema = fields.Schema(id=fields.STORED, a=fields.TEXT, b=fields.TEXT)
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=1, a=u"alfa bravo charlie", b=u"echo foxtrot india")
    w.add_document(id=2, a=u"delta bravo charlie", b=u"alfa alfa alfa")
    w.add_document(id=3, a=u"alfa alfa alfa", b=u"echo foxtrot india")
    w.add_document(id=4, a=u"alfa sierra romeo", b=u"alfa tango echo")
    w.add_document(id=5, a=u"bravo charlie delta", b=u"alfa foxtrot india")
    w.add_document(id=6, a=u"alfa alfa echo", b=u"tango tango tango")
    w.add_document(id=7, a=u"alfa bravo echo", b=u"alfa alfa tango")
    w.commit()
    
    def field_booster(fieldname, factor=2.0):
        "Returns a function which will boost the given field in a query tree"
        def booster_fn(obj):
            if obj.is_leaf() and obj.field() == fieldname:
                obj = obj.copy()
                obj.boost *= factor
                return obj
            else:
                return obj
        return booster_fn
    
    with ix.searcher() as s:
        q = Or([Term("a", u"alfa"), Term("b", u"alfa")])
        
        q = q.accept(field_booster("a", 10.0))
        r = s.search(q)
        for hit in r:
            print hit.score, hit["id"]
        assert_equal([hit["id"] for hit in r], [3, 6, 7, 4, 1, 2, 5])
    
def test_andmaybe_quality():
    schema = fields.Schema(id=fields.STORED, title=fields.TEXT(stored=True),
                           year=fields.NUMERIC)
    ix = RamStorage().create_index(schema)
    
    domain = [(u'Alpha Bravo Charlie Delta', 2000),
              (u'Echo Bravo Foxtrot', 2000), (u'Bravo Golf Hotel', 2002),
              (u'Bravo India', 2002), (u'Juliet Kilo Bravo', 2004),
              (u'Lima Bravo Mike', 2004)]
    w = ix.writer()
    for title, year in domain:
        w.add_document(title=title, year=year)
    w.commit()
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("title", ix.schema)
        q = qp.parse(u"title:bravo ANDMAYBE year:2004")
        
        titles = [hit["title"] for hit in s.search(q, limit=None)[:2]]
        print "titles1=", titles
        assert "Juliet Kilo Bravo" in titles
        
        titles = [hit["title"] for hit in s.search(q, limit=2)]
        print "titles2=", titles
        assert "Juliet Kilo Bravo" in titles

        



















